@@ -33,13 +33,11 @@
      (:temp loop-index unsigned-reg r9-offset))
 
   ;; Pick off the cases where everything fits in register args.
-  (inst test rcx rcx)
-  (inst jmp :z ZERO-VALUES)
-  (inst cmp rcx (fixnumize 1))
+  (inst cmp :dword rcx (fixnumize 1))
   (inst jmp :e ONE-VALUE)
-  (inst cmp rcx (fixnumize 2))
-  (inst jmp :e TWO-VALUES)
-  (inst cmp rcx (fixnumize 3))
+  (inst jmp :b ZERO-VALUES)
+  (inst cmp :dword rcx (fixnumize 3))
+  (inst jmp :b TWO-VALUES)
   (inst jmp :e THREE-VALUES)
 
   ;; As per the calling convention RBX is expected to point at the SP
@@ -74,38 +72,33 @@
   (loadw rsi rbx -3)
 
   ;; And back we go.
-  (inst stc)
-  (inst push rax)
-  (inst ret)
+  (emit-mv-return rax t)
 
   ;; Handle the register arg cases.
   ZERO-VALUES
   (inst lea rbx (ea (* sp->fp-offset n-word-bytes) rbp-tn))
-  (inst mov rdx nil-value)
+  (inst mov rdx null-tn)
   (inst mov rdi rdx)
   (inst mov rsi rdx)
-  (inst mov rsp-tn rbp-tn)
   (inst stc)
-  (inst pop rbp-tn)
+  (inst leave)
   (inst ret)
 
   ;; Note: we can get this, because the return-multiple vop doesn't
   ;; check for this case when size > speed.
   ONE-VALUE
   (loadw rdx rsi -1)
-  (inst mov rsp-tn rbp-tn)
   (inst clc)
-  (inst pop rbp-tn)
+  (inst leave)
   (inst ret)
 
   TWO-VALUES
   (inst lea rbx (ea (* sp->fp-offset n-word-bytes) rbp-tn))
   (loadw rdx rsi -1)
   (loadw rdi rsi -2)
-  (inst mov rsi nil-value)
-  (inst mov rsp-tn rbp-tn)
+  (inst mov rsi null-tn)
   (inst stc)
-  (inst pop rbp-tn)
+  (inst leave)
   (inst ret)
 
   THREE-VALUES
@@ -113,9 +106,8 @@
   (loadw rdx rsi -1)
   (loadw rdi rsi -2)
   (loadw rsi rsi -3)
-  (inst mov rsp-tn rbp-tn)
   (inst stc)
-  (inst pop rbp-tn)
+  (inst leave)
   (inst ret))
 
 ;;;; TAIL-CALL-VARIABLE
@@ -131,7 +123,7 @@
 ;;; we actually called. We also have to compute RCX from the difference
 ;;; between RSI and the stack top.
 #-sb-assembling ; avoid "Redefinition" warning (this file is processed twice)
-(defun !prepare-for-tail-call-variable (fun temp nargs rdx rdi rsi
+(defun prepare-for-tail-call-variable (fun temp nargs rdx rdi rsi
                                         r8 r9 r10
                                         &optional jump-to-the-end)
   (assemble ()
@@ -174,7 +166,7 @@
     ;; And jump into the function.
     (if jump-to-the-end
         (inst jmp end)
-        (inst jmp (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag) fun)))
+        (inst jmp (object-slot-ea fun closure-fun-slot fun-pointer-lowtag)))
 
     ;; All the arguments fit in registers, so load them.
     REGISTER-ARGS
@@ -202,9 +194,9 @@
      (:temp r8 unsigned-reg r8-offset)
      (:temp r9 unsigned-reg r9-offset)
      (:temp r10 unsigned-reg r10-offset))
-  (!prepare-for-tail-call-variable fun temp nargs rdx rdi rsi r8 r9 r10)
+  (prepare-for-tail-call-variable fun temp nargs rdx rdi rsi r8 r9 r10)
 
-  (inst jmp (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag) fun)))
+  (inst jmp (object-slot-ea fun closure-fun-slot fun-pointer-lowtag)))
 
 #+sb-assembling
 (define-assembly-routine
@@ -219,85 +211,77 @@
      (:temp r8 unsigned-reg r8-offset)
      (:temp r9 unsigned-reg r9-offset)
      (:temp r10 unsigned-reg r10-offset))
-  (!prepare-for-tail-call-variable fun temp nargs rdx rdi rsi r8 r9 r10 t)
+  (prepare-for-tail-call-variable fun temp nargs rdx rdi rsi r8 r9 r10 t)
 
   (%lea-for-lowtag-test rbx-tn fun fun-pointer-lowtag)
   (inst test :byte rbx-tn lowtag-mask)
   (inst jmp :nz (make-fixup 'call-symbol :assembly-routine))
-  (inst jmp (ea (- (* closure-fun-slot n-word-bytes) fun-pointer-lowtag) fun)))
+  (inst jmp (object-slot-ea fun closure-fun-slot fun-pointer-lowtag)))
 
 #+sb-assembling
 (define-assembly-routine (call-symbol
+                          (:export undefined-tramp)
                           (:return-style :none))
     ((:temp fun (any-reg descriptor-reg) rax-offset) ; FUN = the symbol
-     (:temp fdefn (any-reg descriptor-reg) rbx-offset))
-  (%lea-for-lowtag-test fdefn fun other-pointer-lowtag)
-  (inst test :byte fdefn lowtag-mask)
+     (:temp tmp (any-reg descriptor-reg) rbx-offset))
+  RETRY
+  (%lea-for-lowtag-test tmp fun other-pointer-lowtag) ; test SYMBOLP
+  (inst test :byte tmp lowtag-mask)
   (inst jmp :nz not-callable)
   (inst cmp :byte (ea (- other-pointer-lowtag) fun) symbol-widetag)
   (inst jmp :ne not-callable)
 
-  (loadw fdefn fun symbol-fdefn-slot other-pointer-lowtag)
-  (inst test :dword fdefn fdefn)
-  (inst jmp :z UNDEFINED)
-  ;; I think we need this MOV because the undefined-function trap examines RAX
-  ;; to see what FDEFN you tried to call through.
-  (inst mov fun fdefn)
-  (inst jmp (ea (- (* fdefn-raw-addr-slot n-word-bytes) other-pointer-lowtag) fdefn))
-  UNDEFINED
-  (inst jmp (make-fixup 'undefined-tramp :assembly-routine))
+  (loadw tmp fun symbol-fdefn-slot other-pointer-lowtag)
+  (inst test :byte tmp tmp) ; if fdefn is nullptr, it has no lowtag
+  (inst jmp :z UNDEFINED-TRAMP)
+  (inst mov fun tmp) ; needed if the JMP invokes UNDEFINED-TRAMP
+  (inst jmp (object-slot-ea tmp closure-fun-slot fun-pointer-lowtag))
   NOT-CALLABLE
-  (inst cmp fun nil-value) ;; NIL doesn't have SYMBOL-WIDETAG
-  (inst jmp :e undefined)
-
+  (inst cmp fun null-tn) ;; NIL doesn't have SYMBOL-WIDETAG
+  (inst jmp :e UNDEFINED-TRAMP)
+  ;; Not a symbol
   (inst pop (ea n-word-bytes rbp-tn))
-  (emit-error-break nil error-trap (error-number-or-lose 'sb-kernel::object-not-callable-error)
-                    (list fun)))
+  (cerror-call nil 'sb-kernel::object-not-callable-error fun)
+  (inst push (ea n-word-bytes rbp-tn))
+  (%lea-for-lowtag-test tmp fun fun-pointer-lowtag) ; test FUNCTIONP
+  (inst test :byte tmp lowtag-mask)
+  (inst jmp :nz RETRY)
+  (inst jmp (object-slot-ea fun closure-fun-slot fun-pointer-lowtag))
+  (inst .align 3)
+  UNDEFINED-TRAMP
+  (inst pop (ea n-word-bytes rbp-tn))
+  (emit-error-break nil cerror-trap (error-number-or-lose 'undefined-fun-error) (list fun))
+  (inst push (ea n-word-bytes rbp-tn))
+  (inst jmp (object-slot-ea fun closure-fun-slot fun-pointer-lowtag)))
 
 
 (define-assembly-routine (throw
-                          (:return-style :raw))
-                         ((:arg target (descriptor-reg any-reg) rdx-offset)
-                          (:arg start any-reg rbx-offset)
-                          (:arg count any-reg rcx-offset)
-                          (:temp bsp-temp any-reg r11-offset)
-                          (:temp catch any-reg rax-offset))
+                             (:return-style :full-call-no-return)
+                           (:save-p :compute-only))
+    ((:arg target (descriptor-reg any-reg) rdx-offset)
+     (:arg start any-reg rbx-offset)
+     (:arg count any-reg rcx-offset)
+     (:temp bsp-temp any-reg r11-offset)
+     (:temp catch any-reg rax-offset))
 
   (declare (ignore start count bsp-temp))
 
   (load-tl-symbol-value catch *current-catch-block*)
-
   LOOP
 
-  (let ((error (gen-label)))
-    (assemble (:elsewhere)
-      (emit-label error)
-
-      ;; Fake up a stack frame so that backtraces come out right.
-      (inst push rbp-tn)
-      (inst mov rbp-tn rsp-tn)
-
-      (emit-error-break nil error-trap
-                        (error-number-or-lose 'unseen-throw-tag-error)
-                        (list target)))
-    (inst test catch catch)             ; check for NULL pointer
-    (inst jmp :z error))
+  (inst test catch catch)               ; check for NULL pointer
+  (inst jmp :z (generate-error-code nil 'unseen-throw-tag-error target))
 
   (inst cmp target (object-slot-ea catch catch-block-tag-slot 0))
-  (inst jmp :e EXIT)
+  ;; Here RAX points to catch block containing symbol pointed to by RDX.
+  (inst jmp :e (make-fixup 'unwind :assembly-routine))
 
   (loadw catch catch catch-block-previous-catch-slot)
-  (inst jmp LOOP)
-
-  EXIT
-
-  ;; Here RAX points to catch block containing symbol pointed to by RDX.
-  ;; An extra RET gets stuffed after the JMP, but oh well. You can't just change
-  ;; the :return-style to :none because that also affects the call sequence.
-  (inst jmp (make-fixup 'unwind :assembly-routine)))
+  (inst jmp LOOP))
 
 ;;; Simply return and enter the loop in UNWIND instead of calling
 ;;; UNWIND directly
+#-sb-assembling
 (define-vop ()
   (:translate %continue-unwind)
   (:policy :fast-safe)
@@ -307,7 +291,8 @@
 (define-assembly-routine (unwind
                           (:return-style :none)
                           (:translate %unwind)
-                          (:policy :fast-safe))
+                          (:policy :fast-safe)
+                          (:save-p :compute-only))
                          ((:arg block (any-reg descriptor-reg) rax-offset)
                           (:arg start (any-reg descriptor-reg) rbx-offset)
                           (:arg count (any-reg descriptor-reg) rcx-offset)
@@ -319,9 +304,9 @@
                           (:temp bsp-temp unsigned-reg r11-offset)
                           (:temp zero complex-double-reg float0-offset))
   AGAIN
-  (let ((error (generate-error-code nil 'invalid-unwind-error)))
-    (inst test block block)             ; check for NULL pointer
-    (inst jmp :z error))
+
+  (inst test block block)               ; check for NULL pointer
+  (inst jmp :z (generate-error-code nil 'invalid-unwind-error))
 
   (load-tl-symbol-value uwp *current-unwind-protect-block*)
 
@@ -356,7 +341,7 @@
 
   ;; Go to uwp-entry, it can fetch the pushed above values if it needs
   ;; to.
-  (inst call (ea (* unwind-block-entry-pc-slot n-word-bytes) uwp))
+  (inst call (object-slot-ea uwp unwind-block-entry-pc-slot 0))
   (inst pop count)
   (inst pop start)
   (inst pop block)
@@ -372,69 +357,7 @@
   (store-tl-symbol-value uwp *current-catch-block*)
 
   ;; nlx-entry expects start in RBX and count in RCX
-  (inst jmp (ea (* unwind-block-entry-pc-slot n-word-bytes) block)))
-
-#+sb-assembling
-(defun ensure-thread-base-tn-loaded ()
-  #-sb-thread
-  (let ((fixup (make-fixup "all_threads" :foreign-dataref)))
-    ;; Load THREAD-BASE-TN from the all_threads. Does not need to be spilled
-    ;; to stack, because we do do not give the register allocator access to it.
-    ;; And call_into_lisp saves it as per convention, not that it matters,
-    ;; because there's no way to get back into C code anyhow.
-    #-immobile-space (inst mov thread-tn (ea fixup))
-    #+immobile-space (inst mov thread-tn (rip-relative-ea fixup))
-    (inst mov thread-tn (ea thread-tn))))
-
-;;; Perform a store to code, updating the GC card mark bit.
-;;; This has two additional complications beyond the ordinary
-;;; generational barrier:
-;;; 1. immobile code uses its own card table which maps linearly
-;;;    with the page index, unlike the dynamic space card table
-;;;    that has a different way of computing a card address.
-;;; 2. code objects are so seldom written that it behooves us to
-;;;    track within each object whether it has been written,
-;;;    thereby avoiding scanning of unwritten objects.
-;;;    This is especially important for immobile space where
-;;;    it is likely that new code will be co-located on a page
-;;;    with old code due to the non-moving allocator.
-#+sb-assembling
-(define-assembly-routine (code-header-set (:return-style :none)) ()
-  ;; stack: ret-pc, object, index, value-to-store
-  (symbol-macrolet ((object (ea 8 rsp-tn))
-                    (word-index (ea 16 rsp-tn))
-                    (newval (ea 24 rsp-tn))
-                    ;; these are declared as vop temporaries
-                    (rax rax-tn)
-                    (rdx rdx-tn)
-                    (rdi rdi-tn))
-  (ensure-thread-base-tn-loaded)
-  (pseudo-atomic ()
-    (assemble ()
-      #+immobile-space
-      (progn
-        (inst mov rax object)
-        (inst sub rax (thread-slot-ea thread-text-space-addr-slot))
-        (inst shr rax (1- (integer-length immobile-card-bytes)))
-        (inst cmp rax (thread-slot-ea thread-text-card-count-slot))
-        (inst jmp :ae try-dynamic-space)
-        (inst mov rdi (thread-slot-ea thread-text-card-marks-slot))
-        (inst bts :dword :lock (ea rdi-tn) rax)
-        (inst jmp store))
-      TRY-DYNAMIC-SPACE
-      (inst mov rax object)
-      (inst shr rax gencgc-card-shift)
-      (inst and :dword rax card-index-mask)
-      (inst mov :byte (ea gc-card-table-reg-tn rax) 0)
-      STORE
-      (inst mov rdi object)
-      (inst mov rdx word-index)
-      (inst mov rax newval)
-      ;; set 'written' flag in the code header
-      (inst or :byte :lock (ea (- 3 other-pointer-lowtag) rdi) #x40)
-      ;; store newval into object
-      (inst mov (ea (- other-pointer-lowtag) rdi rdx n-word-bytes) rax))))
-  (inst ret 24)) ; remove 3 stack args
+  (inst jmp (object-slot-ea block unwind-block-entry-pc-slot 0)))
 
 ;;; These are trampolines, but they benefit from not being in the 'tramps' file
 ;;; because they'll automatically get a vop and an assembly routine this way,
@@ -447,29 +370,199 @@
      (:res r (descriptor-reg) rdx-offset))
   (progn x r)
   (with-registers-preserved (lisp :except rdx)
-    (call-static-fun 'update-object-layout 1)))
+    (call-static-fun 'update-object-layout 1 nil)))
 
-(define-assembly-routine (sb-impl::install-hash-table-lock
+(define-assembly-routine (sb-impl:install-hash-table-lock
                           (:policy :fast-safe)
-                          (:translate sb-impl::install-hash-table-lock)
+                          (:translate sb-impl:install-hash-table-lock)
                           (:return-style :raw))
     ((:arg x (descriptor-reg) rdx-offset)
      (:res r (descriptor-reg) rdx-offset))
   (progn x r)
   (with-registers-preserved (lisp :except rdx)
-    (call-static-fun 'sb-impl::install-hash-table-lock 1)))
+    (call-static-fun 'sb-impl:install-hash-table-lock 1)))
 
-;;; From a perspective of reducing code bloat, this asm routine does not merit
-;;; being one at all. A vop would suffice, because it's a single-use vop.
-;;; However the WITH-REGISTERS-PRESERVED macro seems to think that it is utilized
-;;; only by asm code in *ASSEMBLER-ROUTINES*. I had trouble with it otherwise.
-(define-assembly-routine (alloc-code (:return-style :raw))
-    ((:arg c-arg1 (signed-reg) rdi-offset)
-     (:arg c-arg2 (signed-reg) rsi-offset)
-     (:res res (descriptor-reg) rdx-offset))
-  (progn c-arg1 c-arg2) ; "use" the args
-  ;; Don't preserve the register which holds the lisp return value
-  (with-registers-preserved (c :except rdx)
-    (pseudo-atomic ()
-      (inst call (make-fixup "alloc_code_object" :foreign)))
-    (move res rax-tn)))
+(define-assembly-routine
+    (return-values-list (:return-style :none))
+    ((:arg list descriptor-reg rax-offset)
+
+     (:temp rbx unsigned-reg rbx-offset)
+     (:temp rdx unsigned-reg rdx-offset)
+     (:temp rdi unsigned-reg rdi-offset)
+     (:temp rsi unsigned-reg rsi-offset)
+     (:temp count unsigned-reg rcx-offset)
+     (:temp temp unsigned-reg r9-offset)
+     (:temp return unsigned-reg r10-offset))
+  (symbol-macrolet ((null null-tn))
+  (flet ((check (label)
+           (assemble ()
+             (%test-lowtag list temp skip nil list-pointer-lowtag)
+             (cerror-call nil 'bogus-arg-to-values-list-error list)
+             (inst jmp label)
+             skip)))
+    (assemble ()
+      (%test-lowtag list temp ZERO-VALUES-ERROR t list-pointer-lowtag)
+      (inst cmp list null)
+      (inst jmp :e ZERO-VALUES)
+
+      (loadw rdx list cons-car-slot list-pointer-lowtag)
+      (loadw list list cons-cdr-slot list-pointer-lowtag)
+      (inst cmp list null)
+      (inst jmp :ne CONTINUE)
+      ONE-VALUE
+      (inst clc)
+      (inst leave)
+      (inst ret)
+
+      CONTINUE
+      (check ONE-VALUE)
+
+      (inst mov count (fixnumize 2))
+      (loadw rdi list cons-car-slot list-pointer-lowtag)
+      (loadw list list cons-cdr-slot list-pointer-lowtag)
+      (inst cmp list null)
+      (inst jmp :e TWO-VALUES)
+      (check TWO-VALUES)
+
+      (inst mov count (fixnumize 3))
+      (loadw rsi list cons-car-slot list-pointer-lowtag)
+      (loadw list list cons-cdr-slot list-pointer-lowtag)
+      (inst cmp list null)
+      (inst jmp :e THREE-VALUES)
+      (check THREE-VALUES)
+
+      ;; As per the calling convention RBX is expected to point at the SP
+      ;; before the stack frame.
+      (inst lea rbx (ea (* sp->fp-offset n-word-bytes) rbp-tn))
+
+      (inst mov return (ea (frame-byte-offset return-pc-save-offset) rbp-tn))
+      (inst lea rsp-tn (ea (frame-byte-offset sp->fp-offset) rbp-tn))
+      (inst mov rbp-tn (ea (frame-byte-offset ocfp-save-offset) rbp-tn))
+
+      LOOP
+      (inst add count (fixnumize 1))
+      (pushw list cons-car-slot list-pointer-lowtag)
+      (loadw list list cons-cdr-slot list-pointer-lowtag)
+      (check DONE)
+      (inst cmp list null-tn)
+      (inst jmp :ne LOOP)
+
+      DONE
+      (emit-mv-return return t)
+
+      ZERO-VALUES-ERROR
+      (cerror-call nil 'bogus-arg-to-values-list-error list)
+      ZERO-VALUES
+      (zeroize count)
+      (inst mov rdx null)
+      (inst mov rdi null)
+
+      TWO-VALUES
+      (inst mov rsi null)
+
+      THREE-VALUES
+      (inst lea rbx (ea (* sp->fp-offset n-word-bytes) rbp-tn))
+      (inst stc)
+      (inst leave)
+      (inst ret)))))
+
+;; Adding to the thread-local remset has to be pseudo-atomic because GC takes
+;; ownership of the the vector when it inserts rememberd objects into the common
+;; remset and it assigns 0 into the thread slot.
+;;
+;; For now, a CALL to this routine need not be inside a pseudo-atomic section
+;; because objects are _never_ _removed_ from the permgen remembered set.
+;; This wouldn't be the case if we use the same remembered sets for immobile-space.
+;; If the GC were to remove objects that lack old->young pointers, then the store
+;; would indeed have to be pseudo-atomic with respect to the bit test.
+;;
+(eval-when (:load-toplevel :execute)
+  ;; I want to try different parameters using 'slam' on this file
+  ;; without getting complaints about uneql constants
+  (defparameter remset-vector-capacity 124)
+  (defparameter remset-vector-len (fixnumize (+ remset-vector-capacity 2)))
+  (defparameter remset-vector-words (+ remset-vector-len vector-data-offset)))
+#+(and permgen sb-assembling)
+(progn
+(define-assembly-routine (remset-insert (:return-style :none)) ()
+  (let ((insert (gen-label)))
+  (assemble ()
+  (inst push rbx-tn)
+  ;; Stack:
+  ;;     object
+  ;;     return PC
+  ;;     RAX spill
+  ;;     RBX spill <- sp now
+  ;;     RCX spill
+  ;; Filter out non-permgen objects. RAX is the object
+  ;; (inst mov rbx-tn (rip-relative-ea (make-fixup "permgen_bounds" :foreign-dataref)))
+  (inst mov rbx-tn (ea (make-fixup "permgen_bounds" :foreign-dataref) null-tn))
+  (inst cmp rax-tn (ea rbx-tn))
+  (inst jmp :b SET-BIT-AND-DONE)
+  (inst cmp rax-tn (ea 8 rbx-tn))
+  (inst jmp :ae SET-BIT-AND-DONE)
+  (inst push rcx-tn)
+  (pseudo-atomic ()
+   (inst lfence) ; Don't speculate load of RBX prior to setting P-A
+   (inst mov rbx-tn (thread-slot-ea thread-remset-slot)) ; current remset into RBX
+   (inst test :dword rbx-tn rbx-tn)
+   (inst jmp :z GROW)
+   ;; vector element 0 = item count (max. REMSET-VECTOR-CAPACITY)
+   (inst mov :dword rcx-tn (object-slot-ea rbx-tn vector-data-offset other-pointer-lowtag))
+   (inst cmp :byte rcx-tn (fixnumize remset-vector-capacity))
+   (inst jmp :e GROW)
+   (emit-label INSERT)
+   ;; Check that another thread didn't also insert into its remset.
+   ;; Alternatively, GC could de-duplicate though it currently does not.
+   (inst test :byte rax-tn #b1000) ; SYMBOLP if bit 3
+   (inst jmp :z INSTANCEP)
+   ;; it's a non-NIL symbol
+   (inst bts :lock :dword (ea (- other-pointer-lowtag) rax-tn) 31)
+   (inst jmp MAYBE-SKIP)
+   INSTANCEP
+   (inst bts :lock :dword (ea (- instance-pointer-lowtag) rax-tn) 31)
+   MAYBE-SKIP (inst jmp :c DONE-PA)
+   (inst add :byte (object-slot-ea rbx-tn vector-data-offset other-pointer-lowtag)
+         (fixnumize 1))
+   (inst mov (ea (- (ash (+ 2 vector-data-offset) word-shift) other-pointer-lowtag)
+                 rbx-tn rcx-tn 4)
+         rax-tn) ; RAX=object, RBX=vector, RCX=index(fixnum)
+   DONE-PA)
+  (inst pop rcx-tn)
+  DONE
+  (inst pop rbx-tn) (inst pop rax-tn)
+  (inst ret 8) ; remove arg
+  SET-BIT-AND-DONE
+  ;; layouts and symbol that are not below the core permgen limit reach here.
+  ;; Mark them as remembered to prevent further calls to the slow path.
+  (inst and rax-tn #x-10)
+  (inst or :lock :byte (ea 3 rax-tn) #x80)
+  (inst jmp DONE)
+  GROW
+  ;; if temps regs were needed here, RAX or RBX would be fine to use and reload
+  (emit-allocation nil thread-tn simple-vector-widetag (ash remset-vector-words word-shift)
+                   ;; lowtag, result-tn, allocation-temp
+                   0 rcx-tn nil :systemp t)
+  ;; ASSUMPTION: pre-zeroed memory
+  (inst mov :byte (object-slot-ea rcx-tn 0 0) simple-vector-widetag) ; header
+  (inst mov :byte (object-slot-ea rcx-tn vector-length-slot 0) remset-vector-len)
+  (inst mov (object-slot-ea rcx-tn (1+ vector-data-offset) 0) rbx-tn) ; link buffers
+  (inst lea rbx-tn (ea other-pointer-lowtag rcx-tn)) ; tagify
+  (inst mov (thread-slot-ea thread-remset-slot) rbx-tn) ; writeback
+  (zeroize rcx-tn)
+  (inst jmp INSERT))))
+(define-assembly-routine (gc-remember-symbol (:return-style :none)) ()
+  (inst push rax-tn)
+  (inst mov rax-tn (ea (ash 2 word-shift) rsp-tn)) ; OBJECT
+  (inst test :byte (ea (- 3 other-pointer-lowtag) rax-tn) #x80)
+  (inst jmp :z (make-fixup 'remset-insert :assembly-routine))
+  (inst pop rax-tn)
+  (inst ret 8))
+(define-assembly-routine (gc-remember-layout (:return-style :none)) ()
+  (inst push rax-tn)
+  (inst mov rax-tn (ea (ash 2 word-shift) rsp-tn)) ; OBJECT
+  (inst test :byte (ea (- 3 instance-pointer-lowtag) rax-tn) #x80)
+  (inst jmp :z (make-fixup 'remset-insert :assembly-routine))
+  (inst pop rax-tn)
+  (inst ret 8))
+)
