@@ -59,7 +59,7 @@
             (error "Won't do (RATIONALIZE ~S) due to possible precision loss" x)))))
 
 (defun xfloat-coerce (object type)
-  (declare (number object))
+  (declare (type number object))
   (when (member type '(integer rational real))
     ;; This branch won't accept (coerce x 'real) if X is one of our
     ;; target-floats. We don't need that apparently.
@@ -398,50 +398,47 @@
       (0d0 1d0)
       (t (error "Unimplemented.")))))
 
+(defun natural-log (number)
+  (declare (rational number))
+  (flet ((%log (y)
+           (declare (rational y))
+           (assert (<= 1/2 y 1))
+           ;; Use the Taylor series expansion for log((1-y)/(1+y)),
+           ;; which gives around 4*m bits of precision for m terms
+           ;; when 1/2 <= y <= 1.
+           (let* ((x (/ (- y 1) (+ y 1)))
+                  (x^2 (* x x))
+                  (sum 0))
+             (do ((k 1 (+ k 2))
+                  (x^k x (* x^2 x^k)))
+                 ;; 40 * 4 = 160 bits of precision.
+                 ((>= k 40) (* 2 sum))
+               (incf sum (/ x^k k))))))
+    ;; Write NUMBER as 2^k * a, so that 1/2 < a <= 1. Then log(NUMBER)
+    ;; is log(a) - k * log(1/2).
+    (let* ((k (1+ (- (integer-length (numerator number))
+                     (integer-length (denominator number)))))
+           (a (/ number (expt 2 k))))
+      (- (%log a) (* k (%log 1/2))))))
+
 (defun log (number &optional (base nil base-p))
   (validate-args number base)
+  (when (eql number 0)
+    (error 'division-by-zero :operation 'log :operands `(,number ,@(when base-p base))))
   (with-memoized-math-op (log (if base-p (list number base) number))
     (let ((format (pick-float-result-format number (if base-p base 0))))
       (if (zerop number)
           (coerce single-float-negative-infinity format)
-          (case base
-            ((nil)
-             (let ((table '((1 . 0f0)
-                            (10 . 2.3025851f0)
-                            (#x1fffffff . 20.101269f0)
-                            (#x20000000 . 20.101269f0)
-                            (#x20000001 . 20.101269f0)
-                            (#xfffffffffffffff . 41.58883f0)
-                            (#x1000000000000000 . 41.58883f0)
-                            (#x1000000000000001 . 41.58883f0)
-                            (#x3fffffffffffffff . 42.975124f0)
-                            (#x4000000000000000 . 42.975124f0)
-                            (#x4000000000000001 . 42.975124f0)
-                            (0.9999999999999999d0 . -1.1102230246251565d-16)
-                            (1d0 . 0d0)
-                            (2d0 . 0.6931471805599453d0)
-                            (2.718281828459045d0 . 1d0)
-                            (5.36870911d8 . 20.10126823437577d0)
-                            (5.36870912d8 . 20.101268236238415d0)
-                            (2.147483647d9 . 21.487562596892644d0)
-                            (2.147483648d9 . 21.487562597358306d0)
-                            (1.152921504606847d18 . 41.58883083359672d0)
-                            (4.611686018427388d18 . 42.97512519471661d0)
-                            (9.223372036854776d18 . 43.66827237527655d0))))
-               (or (cdr (assoc number table))
-                   (error "missing entry for (LOG ~A)" number))))
-            ((10 10f0 10d0)
-             (let ((table '((2d0 . 0.3010299956639812d0))))
-               (or (cdr (assoc number table))
-                   (error "missing entry for (LOG ~A 10)" number))))
-            (t (error "missing entries for (LOG ~A ~A)" number base)))))))
+          (flonum-from-rational (/ (natural-log (rational number))
+                                   (if base-p
+                                       (natural-log (rational base))
+                                       1))
+                                format)))))
 
-;;; Canonicalize and write out the memoization table. Order it by function,
-;;; then number of arguments, then each argument's value. Rational precedes
-;;; single-float which precedes double-float, then order by bits.
+;;; Canonicalize and write out the memoization table.
 (defun dump-math-memoization-table (table stream)
   (format stream ";;; This file is machine-generated. DO NOT EDIT~2%")
-  (format stream ":DEFAULT~%(~%")
+  (format stream "~%(~%")
   (labels ((spelling-of (expr)
              ;; MUST not write package prefixes !
              ;; e.g. avoid writing a line like (COERCE (-33619991 SB-XC:DOUBLE-FLOAT) ...)
