@@ -214,6 +214,14 @@
   ;; for classes named by a symbol, otherwise a pseudo-random value.
   ;; Must be acceptable as an argument to SB-INT:MIX
   (clos-hash (missing-arg) :type (and fixnum unsigned-byte))
+  ;; Vtable could be used to store the per-layout implementation of the combined methods for
+  ;; any single-dispatch function. The trick is to dynamically figure out when it would make
+  ;; sense to claim a unique index into the vtable. Certainly for a function that has about half
+  ;; as many methods as there are classes e.g. #<STANDARD-GENERIC-FUNCTION CL-PROTOBUFS:CLEAR (2017)>
+  ;; when there are 4000 classes. A PCL cache (key = layout, value = method) would take up double
+  ;; the space at least, plus all the unused cells required to make the probing strategy work.
+  ;; And it wouldn't be as quick as a direct lookup.
+  (vtable)
   ;; the class that this is a layout for
   (classoid (missing-arg) :type classoid)
   ;; The value of this slot can be:
@@ -575,7 +583,9 @@
 (defun copy-ctype (x &optional (flags 0))
   (declare (type ctype x))
   (declare (sb-c::tlab :system) (inline !new-xset))
-  #+c-stack-is-control-stack (aver (stack-allocated-p x))
+  #.(cl:if (cl:and (cl:member :c-stack-is-control-stack sb-xc:*features*)
+                   sb-ext:*stack-allocate-dynamic-extent*)
+           '(aver (stack-allocated-p x)))
   (labels ((copy (x)
              ;; Return a heap copy of X if X was arena or stack-allocated.
              ;; I suspect it's quicker to copy always rather than conditionally.
@@ -628,9 +638,9 @@
                       (fun-type-wild-args x) (fun-type-returns x))))
            (%set-instance-layout copy (%instance-layout x))
            copy))
-        (numeric-type
-         (!alloc-numeric-type bits (numeric-type-aspects x)
-                              (copy (numeric-type-low x)) (copy (numeric-type-high x))))
+        (numeric-union-type
+         (!alloc-numeric-union-type bits (numeric-union-type-aspects x)
+                                    (map 'vector #'copy (numeric-union-type-ranges x))))
         (compound-type ; UNION or INTERSECTION
          (let ((copy (!alloc-union-type bits (compound-type-enumerable x)
                                         (compound-type-types x))))
@@ -661,7 +671,7 @@
 
 #-sb-xc-host
 (progn
-(defglobal *!initial-ctypes* nil)
+(define-load-time-global *!initial-ctypes* nil)
 (defun preload-ctype-hashsets ()
   (dolist (pair (nreverse *!initial-ctypes*))
     (let ((instance (car pair))
@@ -694,7 +704,7 @@
                        (bug "genesis dumped bad instance within ~X"
                             (get-lisp-obj-address instance)))))))
         (etypecase instance
-          ((or numeric-type member-type character-set-type ; nothing extra to do
+          ((or numeric-union-type member-type character-set-type ; nothing extra to do
            #+sb-simd-pack simd-pack-type #+sb-simd-pack-256 simd-pack-256-type
            hairy-type))
           (args-type
