@@ -17,27 +17,15 @@
 
 ;;;; fdefinition (fdefn) objects
 
-(defun make-fdefn (name)
-  #-(and x86-64 immobile-space) (make-fdefn name)
-  #+(and x86-64 immobile-space)
-  (let ((fdefn (truly-the (values fdefn &optional)
-                          (sb-vm::alloc-immobile-fdefn))))
-    (%primitive sb-vm::set-slot fdefn name 'make-fdefn
-                sb-vm:fdefn-name-slot sb-vm:other-pointer-lowtag)
-    (fdefn-makunbound fdefn)
-    fdefn))
-
-(defun undo-static-linkage (fdefn) (declare (ignore fdefn)))
-
+(declaim (ftype (sfunction (t) (or fdefn #+linkage-space symbol)) find-or-create-fdefn))
+#-linkage-space
+(progn
+(defun fset (fname fun)
+  (declare (type (or fdefn symbol list) fname))
+  (setf (fdefn-fun (if (fdefn-p fname) fname (find-or-create-fdefn fname))) fun))
 (defun (setf fdefn-fun) (fun fdefn)
-  (declare (type function fun)
-           (type fdefn fdefn))
-  (undo-static-linkage fdefn)
-  (sb-c::when-vop-existsp (:named sb-vm::set-fdefn-fun)
-    (%primitive sb-vm::set-fdefn-fun fun fdefn))
-  (sb-c::unless-vop-existsp (:named sb-vm::set-fdefn-fun)
-    (sb-vm::set-fdefn-fun fun fdefn))
-  fun)
+  (%primitive sb-vm::set-fdefn-fun fun fdefn)
+  fun))
 
 ;;; Return the FDEFN object for NAME, or NIL if there is no fdefn.
 ;;; Signal an error if name isn't valid.
@@ -46,6 +34,9 @@
 (defun find-fdefn (name)
   (declare (explicit-check))
   (when (symbolp name) ; Don't need LEGAL-FUN-NAME-P check
+    ;; FDEFNs can only be created for names that aren't symbols with #+linkage-space
+    #+linkage-space (bug "FIND-FDEFN ~S" name)
+    #-linkage-space
     (let ((fdefn (sb-vm::%symbol-fdefn name))) ; slot default is 0, not NIL
       (return-from find-fdefn (if (eql fdefn 0) nil fdefn))))
   ;; Technically the ALLOW-ATOM argument of NIL isn't needed, but
@@ -104,11 +95,12 @@
                   symbol))
           (t def))))
 
-(define-load-time-global *fdefn-of-nil* (make-fdefn nil))
-(declaim (ftype (sfunction (t) fdefn) find-or-create-fdefn))
+#-linkage-space (define-load-time-global *fdefn-of-nil* (make-fdefn nil))
 (defun find-or-create-fdefn (name)
   (cond
     ((symbolp name)
+     #+linkage-space name
+     #-linkage-space
      (let ((fdefn (sb-vm::%symbol-fdefn name)))
        (cond ((and fdefn (neq fdefn 0)) fdefn)
              ((null name) *fdefn-of-nil*)
@@ -217,7 +209,8 @@
                                (:copier nil))
   (type nil :type symbol)
   ;; the underlying definition prior to getting wrapped in a closure
-  (definition nil :type function))
+  (definition nil :type function)
+  (specialized-xep nil :type (or null function)))
 (declaim (freeze-type encapsulation-info))
 
 ;;; Find the encapsulation info that has been closed over.
@@ -257,24 +250,41 @@
 ;;; otherwise the new encapsulation goes to the front of the chain.
 (defun encapsulate (name type function)
   (let ((underlying-fun (name->fun name)))
-    (when (macro/special-guard-fun-p underlying-fun)
+    ;; No error is signaled when encapsulating a nonexistent function, however
+    ;; the wrapper will receive as its first arg a function that signals an error.
+    (when (or (and underlying-fun (macro/special-guard-fun-p underlying-fun))
+              (not name))
       (error "~S can not be encapsulated" name))
     (when (typep underlying-fun 'generic-function)
       (return-from encapsulate
         (encapsulate-generic-function underlying-fun type function)))
     (multiple-value-bind (existing predecessor) (has-encap underlying-fun type)
       ;; If TYPE existed, the new DEFINITION comes from the existing
-      (when existing
-        (setf underlying-fun (encapsulation-info-definition existing)))
-      (let* ((info (make-encapsulation-info type underlying-fun))
+      (let* ((info (if (not existing)
+                       (make-encapsulation-info
+                        type
+                        (or underlying-fun
+                            (lambda (&rest args)
+                              (declare (sb-c::lambda-list ($undef$)))
+                              (declare (ignore args))
+                              (error "~S is undefined" name))))
+                       (copy-structure existing)))
+             (specialized-xep (info :function :specialized-xep name))
              (closure (named-lambda encapsulation (&rest args)
                         (apply function (encapsulation-info-definition info)
                                args))))
+        (when specialized-xep
+          (let* ((xep-name
+                   `(specialized-xep ,name ,@specialized-xep))
+                 (xep-fun (name->fun xep-name)))
+            (when xep-fun
+              (setf (encapsulation-info-specialized-xep info) xep-fun)
+              (remove-specialized-xep name nil))))
         (if predecessor
             ;; Become the successor of the existing predecessor
             (setf (encapsulation-info-definition predecessor) closure)
             ;; Was first in chain or didn't exist
-            (setf (fdefn-fun (find-fdefn name)) closure))))))
+            (fset name closure))))))
 
 (defun unencapsulate (name type)
   "Removes NAME's encapsulation of the specified TYPE if such exists."
@@ -283,11 +293,15 @@
         (unencapsulate-generic-function fun type)
         (multiple-value-bind (existing predecessor) (has-encap fun type)
           (when existing
-            (let ((next (encapsulation-info-definition existing)))
-              (if predecessor
-                  (setf (encapsulation-info-definition predecessor) next)
-                  ;; It's the first one, so change the fdefn object.
-                  (setf (fdefn-fun (find-fdefn name)) next)))))))))
+            (let* ((next (encapsulation-info-definition existing))
+                   (specialized-xep (encapsulation-info-specialized-xep existing)))
+              (cond (predecessor
+                     (setf (encapsulation-info-definition predecessor) next))
+                    ;; It's the first one, so change the fdefn object.
+                    ((equal (%fun-lambda-list next) '($undef$)) (fmakunbound name))
+                    (t (fset name next)))
+              (when specialized-xep
+                (fset (%fun-name specialized-xep) specialized-xep)))))))))
 
 
 ;;;; FDEFINITION
@@ -349,8 +363,20 @@
            :format-control "~S is not acceptable to ~S."
            :format-arguments (list object setter))))
 
-(defun (setf fdefinition) (new-value name)
-  "Set NAME's global function definition."
+(defun remove-specialized-xep (name &optional (clear-info t))
+  (let ((xep (info :function :specialized-xep name)))
+    (when xep
+      (let* ((xep-name
+               `(specialized-xep ,name ,@xep))
+             (fdefn (find-fdefn xep-name)))
+        (when fdefn
+          (fset fdefn
+                (compile nil (make-specialized-xep-stub name xep xep-name))))
+        (when clear-info
+          (clear-info :function :specialized-xep name)
+          (undefine-fun-name xep-name))))))
+
+(defun setf-fdefinition (new-value name clear-specialized-xep)
   (declare (type function new-value) (optimize (safety 1)))
   (declare (explicit-check))
   (err-if-unacceptable-function new-value '(setf fdefinition))
@@ -363,46 +389,55 @@
     (when (symbolp name)
       (let ((old (%symbol-function name)))
         (dolist (spec *user-hash-table-tests*)
-            (cond ((eq old (second spec))
-                   ;; test-function
-                   (setf (second spec) new-value))
-                  ((eq old (third spec))
-                   ;; hash-function
-                   (setf (third spec) new-value))))))
+          (cond ((eq old (second spec))
+                 ;; test-function
+                 (setf (second spec) new-value))
+                ((eq old (third spec))
+                 ;; hash-function
+                 (setf (third spec) new-value))))))
 
     (let ((fdefn (find-or-create-fdefn name)))
       (dolist (f *setf-fdefinition-hook*)
         (declare (type function f))
         (funcall f name new-value))
+      (when clear-specialized-xep
+        (remove-specialized-xep name))
+      ;; fdefn may be a symbol if #+linkage-space. fdefn-fun vop is fine with that
       (let ((encap-info (encapsulation-info (fdefn-fun fdefn))))
         (cond (encap-info
                (loop
                 (let ((more-info
-                       (encapsulation-info
-                        (encapsulation-info-definition encap-info))))
+                        (encapsulation-info
+                         (encapsulation-info-definition encap-info))))
                   (if more-info
                       (setf encap-info more-info)
                       (return (setf (encapsulation-info-definition encap-info)
                                     new-value))))))
               (t
-               (setf (fdefn-fun fdefn) new-value)))))))
+               (fset fdefn new-value)))))))
+
+(defun (setf fdefinition) (new-value name)
+  "Set NAME's global function definition."
+  (declare (type function new-value) (optimize (safety 1)))
+  (declare (explicit-check))
+  (setf-fdefinition new-value name t))
 
 ;;;; FBOUNDP and FMAKUNBOUND
 
 (defun fboundp (name)
   "Return true if name has a global function definition."
   (declare (explicit-check))
-  (awhen (find-fdefn name) (fdefn-fun it)))
+  #+linkage-space (fdefn-fun (if (symbolp name) name (find-fdefn name)))
+  #-linkage-space (awhen (find-fdefn name) (fdefn-fun it)))
 
 (defun fmakunbound (name)
   "Make NAME have no global function definition."
   (declare (explicit-check))
   (with-single-package-locked-error
       (:symbol name "removing the function or macro definition of ~A")
-    (let ((fdefn (find-fdefn name)))
-      (when fdefn
-        (undo-static-linkage fdefn)
-        (fdefn-makunbound fdefn)))
+    #+linkage-space (acond ((symbolp name) (fset name 0))
+                           ((find-fdefn name) (fset it 0)))
+    #-linkage-space (awhen (find-fdefn name) (fdefn-makunbound it))
     (undefine-fun-name name)
     name))
 
