@@ -82,7 +82,6 @@
   (def set-header-data (x val))
   (def widetag-of)
   (def %other-pointer-widetag)
-  (def pointer-hash)
   (def vector-sap)
   (def binding-stack-pointer-sap  ())
   #+cheneygc (def dynamic-space-free-pointer ())
@@ -98,7 +97,7 @@
   (def sb-c::vector-length)
   (def make-array-header (type rank))
   (def code-instructions)
-  #-untagged-fdefns (def code-header-ref (code-obj index))
+  (def code-header-ref (code-obj index))
   (def %vector-raw-bits (object offset))
   (def %set-vector-raw-bits (object offset value))
   #-weak-vector-readbarrier (def weak-vector-len)
@@ -141,7 +140,7 @@
   (def* (%array-atomic-incf/word (array index diff))
         (%raw-instance-atomic-incf/word (instance index diff)))
   #+(or x86 x86-64)
-  (def* (sb-vm::%cpu-identification (arg1 arg2))
+  (def* (sb-vm::%cpu-identification (eax ecx))
         (sb-vm::%vector-cas-pair (vector index old1 old2 new1 new2))
         (sb-vm::%instance-cas-pair (instance index old1 old2 new1 new2))
         (sb-vm::%cons-cas-pair (cons old1 old2 new1 new2)))
@@ -166,13 +165,14 @@
   (def symbol-package-id)
   (def symbol-hash)
   (def symbol-%info) ; primitive reader always needs a stub
-  #-(or x86 x86-64) (def lra-code-header)
+  #-(or x86 x86-64 arm64 riscv) (def lra-code-header)
   (def %make-lisp-obj)
-  (def get-lisp-obj-address)
   #+x86-64
   (def single-float-copysign (float float2))
   #+x86-64
-  (def single-float-sign))
+  (def single-float-sign)
+  #+64-bit
+  (def %make-double-float))
 
 #+sb-simd-pack
 (macrolet ((def (name)
@@ -214,3 +214,9 @@
   (setq sb-c::*undefined-warnings* nil))
 
 (setf (fdefinition 'unaligned-dx-cons) #'list)
+
+(defun rotate-right-word (n count)
+  (dpb
+   n
+   (byte count (- sb-vm:n-word-bits count))
+   (ash n (- count))))
