@@ -18,6 +18,8 @@
 ;;; when we descend into a compound object or when we step through elements of
 ;;; a compound object.
 (defconstant +max-hash-depthoid+ 4)
+
+(defconstant +max-hash-table-bits+ #-64-bit 29 #+64-bit 31)
 
 ;;; This is an out-of-line callable entrypoint that the compiler can
 ;;; transform SXHASH into when hashing a non-simple string.
@@ -28,10 +30,29 @@
           (with-array-data ((string x) (start) (end) :check-fill-pointer t)
             (values string start end))
           (values x 0 (length x)))
-    ;; I'm not sure I believe the comment about FLET TRICK being needed.
-    ;; The generated code seems tight enough, and the comment is, after all,
-    ;; >14 years old.
     (%sxhash-simple-substring string start end)))
+
+(defun %sxhash-string/truncating (x max-length)
+  (declare (type string x)
+           (type fixnum max-length)
+           (optimize speed (safety 0)))
+  (multiple-value-bind (string start end)
+      (if (array-header-p x)
+          (with-array-data ((string x) (start) (end) :check-fill-pointer t)
+            (values string start end))
+          (values x 0 (length x)))
+    (%sxhash-simple-substring/truncating string start end max-length)))
+
+(defmacro %%sxhash-string (x max-length on-truncate)
+  (once-only ((x x) (max-length max-length))
+    (with-unique-names (hash n-chars-truncated)
+      `(if ,max-length
+           (multiple-value-bind (,hash ,n-chars-truncated)
+               (%sxhash-string/truncating ,x ,max-length)
+             (unless (zerop ,n-chars-truncated)
+               ,on-truncate)
+             ,hash)
+           (%sxhash-string ,x)))))
 
 ;;;; the SXHASH function
 
@@ -58,7 +79,8 @@
 ;;; Since WITH-PINNED-OBJECT costs nothing on conservative gencgc,
 ;;; that's what I'm going with.
 ;;;
-(defun %instance-sxhash (instance)
+(declaim (inline %instance-sxhash))
+(defun %instance-sxhash (instance header-word)
   ;; LAYOUT must not acquire an extra slot for the stable hash,
   ;; because the bitmap length is derived from the instance length.
   ;; It would probably be simple to eliminate this as a special case
@@ -68,28 +90,31 @@
     ;; This might be wrong if the clos-hash was clobbered to 0
     (return-from %instance-sxhash (layout-clos-hash instance)))
   ;; Non-simple cases: no hash slot, and either unhashed or hashed-not-moved.
-  (let* ((header-word (instance-header-word instance))
-         (addr (with-pinned-objects (instance)
-                 ;; First we have to indicate that a hash was taken from the address
-                 ;; if not already so marked.
-                 (unless (logbitp sb-vm:stable-hash-required-flag header-word)
-                   #-sb-thread (setf (sap-ref-word (int-sap (get-lisp-obj-address instance))
-                                                   (- sb-vm:instance-pointer-lowtag))
-                                     (logior (ash 1 sb-vm:stable-hash-required-flag)
-                                             header-word))
-                   #+sb-thread (%primitive sb-vm::set-instance-hashed instance))
-                 (get-lisp-obj-address instance))))
+  (let* ((addr (sb-c::if-vop-existsp (:named sb-vm::set-instance-hashed-return-address)
+                 (if (logbitp sb-vm:stable-hash-required-flag header-word)
+                     (get-lisp-obj-address instance)
+                     (%primitive sb-vm::set-instance-hashed-return-address instance))
+                 (with-pinned-objects (instance)
+                   ;; First we have to indicate that a hash was taken from the address
+                   ;; if not already so marked.
+                   (unless (logbitp sb-vm:stable-hash-required-flag header-word)
+                     #-sb-thread (setf (sap-ref-word (int-sap (get-lisp-obj-address instance))
+                                                     (- sb-vm:instance-pointer-lowtag))
+                                       (logior (ash 1 sb-vm:stable-hash-required-flag)
+                                               header-word))
+                     #+sb-thread (%primitive sb-vm::set-instance-hashed instance))
+                   (get-lisp-obj-address instance)))))
     ;; perturb the address
     (murmur-hash-word/+fixnum addr)))
 
 (declaim (inline instance-sxhash))
 (defun instance-sxhash (instance)
-  (if (logbitp sb-vm:hash-slot-present-flag
-               (instance-header-word (truly-the instance instance)))
-      ;; easy case: 1 word beyond the apparent length is a word added
-      ;; by GC (which may have resized the object, but we don't need to know).
-      (truly-the hash-code (%instance-ref instance (%instance-length instance)))
-      (%instance-sxhash instance)))
+  (let ((header-word (instance-header-word (truly-the instance instance))))
+    (if (logbitp sb-vm:hash-slot-present-flag header-word)
+        ;; easy case: 1 word beyond the apparent length is a word added
+        ;; by GC (which may have resized the object, but we don't need to know).
+        (truly-the hash-code (%instance-ref instance (%instance-length instance)))
+        (%instance-sxhash instance header-word))))
 
 ;;; Return a pseudorandom number that was assigned on allocation.
 ;;; FIN is a STANDARD-FUNCALLABLE-INSTANCE but we don't care to type-check it.
@@ -252,76 +277,108 @@
   ;; toss in a LOGNOT so that (the word a) and (int-sap a) hash differently
   (murmur-hash-word/+fixnum (logand (lognot (sap-int x)) most-positive-word)))
 
-(defun sxhash (x)
-  ;; profiling SXHASH is hard, but we might as well try to make it go
-  ;; fast, in case it is the bottleneck somewhere.  -- CSR, 2003-03-14
-  ;; So, yes, profiling is a little tough but not impossible with some added
-  ;; instrumentation in each stanza of the COND, either manually or
-  ;; automagically. Based on a manual approach, the order of the tests below
-  ;; are now better arranged by approximate descending frequency in terms
-  ;; of calls observed in certain test. Regardless of the fact that applications
-  ;; will vary by use-cases, this seems like a good order because:
-  ;;  * despite that INSTANCE is often the 2nd-most common object type in the heap
-  ;;    (right behind CONS), there are probably at least as many heap words
-  ;;    that are FIXNUM as instance pointers. So it stands to reason that
-  ;;    SXHASH-RECURSE is invoked very often on FIXNUM.
-  ;;  * SYMBOLs are extremely common as table keys, more so than INSTANCE,
-  ;;    so we should pick off SYMBOL sooner than INSTANCE as well.
-  ;;  * INSTANCE (except for PATHNAME) doesn't recurse anyway - in fact
-  ;;    it is particularly dumb (by design), so performing that test later
-  ;;    doesn't incur much of a penalty.
-  ;; Anyway, afaiu, the code below was previously ordered by gut feeling
-  ;; rather than than actual measurement, so having any rationale for ordering
-  ;; is better than having no rationale. And as a further comment observes,
-  ;; we could do away with the question of order if only we had jump tables.
-  ;; (Also, could somebody perhaps explain how these magic numbers were chosen?)
-  (declare (optimize speed))
-  (labels ((sxhash-recurse (x depthoid)
-             (declare (type index depthoid))
-             (typecase x
-               ;; we test for LIST here, rather than CONS, because the
-               ;; type test for CONS is in fact the test for
-               ;; LIST-POINTER-LOWTAG followed by a negated test for
-               ;; NIL.  If we're going to have to test for NIL anyway,
-               ;; we might as well do it explicitly and pick off the
-               ;; answer.  -- CSR, 2004-07-14
-               (list
-                (if (null x)
-                    (sxhash x)          ; through DEFTRANSFORM
-                    (if (plusp depthoid)
+;;; Like SXHASH, but the amount of hashing effort can be controlled.
+;;;
+;;; MAX-DEPTHOID overrides +MAX-HASH-DEPTHOID+.
+;;;
+;;; MAX-LENGTH limits the number of characters to hash in a single
+;;; string. Currently, an equal number of characters from the
+;;; beginning and end of the string are hashed. NIL means no limit.
+;;;
+;;; ON-TRUNCATE is a form that's evaluated whenever (even multiple
+;;; times) some data in X is discarded and higher limits (MAX-ATOMS
+;;; and MAX-LENGTH) may change the hash. Thus ON-TRUNCATE is not
+;;; evaluated if we bail out because there are too many cycles.
+(defmacro %sxhash (x max-depthoid max-length &key on-truncate)
+  (once-only ((x x) (max-depthoid max-depthoid) (max-length max-length))
+    `(locally (declare (optimize speed))
+       ;; Profiling SXHASH is hard, but we might as well try to make
+       ;; it go fast, in case it is the bottleneck somewhere. -- CSR,
+       ;; 2003-03-14 So, yes, profiling is a little tough but not
+       ;; impossible with some added instrumentation in each stanza of
+       ;; the COND, either manually or automagically. Based on a
+       ;; manual approach, the order of the tests below are now better
+       ;; arranged by approximate descending frequency in terms of
+       ;; calls observed in certain test. Regardless of the fact that
+       ;; applications will vary by use-cases, this seems like a good
+       ;; order because:
+       ;;
+       ;;  * despite that INSTANCE is often the 2nd-most common object
+       ;;    type in the heap (right behind CONS), there are probably
+       ;;    at least as many heap words that are FIXNUM as instance
+       ;;    pointers. So it stands to reason that FIXNUMs quite
+       ;;    commonly occur in objects.
+       ;;
+       ;;  * SYMBOLs are extremely common as table keys, more so than
+       ;;    INSTANCE, so we should pick off SYMBOL sooner than
+       ;;    INSTANCE as well.
+       ;;
+       ;;  * INSTANCE (except for PATHNAME) doesn't recurse anyway -
+       ;;    in fact it is particularly dumb (by design), so
+       ;;    performing that test later doesn't incur much of a
+       ;;    penalty.
+       ;;
+       ;; Anyway, afaiu, the code below was previously ordered by gut
+       ;; feeling rather than than actual measurement, so having any
+       ;; rationale for ordering is better than having no rationale.
+       ;; And as a further comment observes, we could do away with the
+       ;; question of order if only we had jump tables. (Also, could
+       ;; somebody perhaps explain how these magic numbers were
+       ;; chosen?)
+       (labels
+           ((sxhash-recurse (x depthoid)
+              (declare (type index depthoid))
+              (typecase x
+                ;; We test for LIST here, rather than CONS, because
+                ;; the type test for CONS is in fact the test for
+                ;; LIST-POINTER-LOWTAG followed by a negated test for
+                ;; NIL. If we're going to have to test for NIL anyway,
+                ;; we might as well do it explicitly and pick off the
+                ;; answer. -- CSR, 2004-07-14
+                (list
+                 (cond ((null x)
+                        (sxhash x))     ; through DEFTRANSFORM
+                       ((plusp depthoid)
                         (mix (sxhash-recurse (car x) (1- depthoid))
-                             (sxhash-recurse (cdr x) (1- depthoid)))
+                             (sxhash-recurse (cdr x) (1- depthoid))))
+                       (t
+                        (when (< ,max-depthoid 32)
+                          ,on-truncate)
                         261835505)))
-               (symbol (sxhash x)) ; through DEFTRANSFORM
-               (fixnum (sxhash x)) ; through DEFTRANSFORM
-               (instance
-                (if (pathnamep x)
-                    (pathname-sxhash x)
-                    (instance-sxhash x)))
-               (array
-                (typecase x
-                  (string (%sxhash-string x))
-                  (bit-vector (%sxhash-bit-vector x))
-                  ;; Would it be legal to mix in the widetag?
-                  (t (logxor 191020317 (sxhash (array-rank x))))))
-               ;; general, inefficient case of NUMBER
-               ;; There's a spurious FIXNUMP test here, as we've already picked it off.
-               ;; Maybe the NUMBERP emitter could be informed that X can't be a fixnum,
-               ;; because writing this case as (OR BIGNUM RATIO FLOAT COMPLEX)
-               ;; produces far worse code.
-               (number (number-sxhash x))
-               (character
-                (logxor 72185131
-                        (sxhash (char-code x)))) ; through DEFTRANSFORM
-               (funcallable-instance
-                (if (logtest (layout-flags (%fun-layout x)) +pcl-object-layout-flag+)
-                    ;; We have a hash code, so might as well use it.
-                    (fsc-instance-hash x)
-                    ;; funcallable structure, not funcallable-standard-object
-                    9550684))
-               (system-area-pointer (sap-hash x))
-               (t 42))))
-    (sxhash-recurse x +max-hash-depthoid+)))
+                (symbol (sxhash x))     ; through DEFTRANSFORM
+                (fixnum (sxhash x))     ; through DEFTRANSFORM
+                (instance
+                 (if (pathnamep x)
+                     (pathname-sxhash x)
+                     (instance-sxhash x)))
+                (array
+                 (typecase x
+                   (string (%%sxhash-string x ,max-length ,on-truncate))
+                   (bit-vector (%sxhash-bit-vector x))
+                   (t
+                    ;; We could even mix in the widetag.
+                    (logxor 191020317 (sxhash (array-rank x))))))
+                ;; general, inefficient case of NUMBER
+                ;; There's a spurious FIXNUMP test here, as we've already picked it off.
+                ;; Maybe the NUMBERP emitter could be informed that X can't be a fixnum,
+                ;; because writing this case as (OR BIGNUM RATIO FLOAT COMPLEX)
+                ;; produces far worse code.
+                (number (number-sxhash x))
+                (character
+                 (logxor 72185131
+                         (sxhash (char-code x)))) ; through DEFTRANSFORM
+                (funcallable-instance
+                 (if (logtest (layout-flags (%fun-layout x)) +pcl-object-layout-flag+)
+                     ;; We have a hash code, so might as well use it.
+                     (fsc-instance-hash x)
+                     ;; funcallable structure, not funcallable-standard-object
+                     9550684))
+                (system-area-pointer (sap-hash x))
+                (t 42))))
+         (sxhash-recurse ,x (min (truly-the index ,max-depthoid) 32))))))
+
+(defun sxhash (x)
+  (%sxhash x +max-hash-depthoid+ nil))
 
 ;;;; the PSXHASH function
 
@@ -432,17 +489,17 @@
                             (mixf result (logand (%raw-instance-ref/word key i)
                                                  most-positive-fixnum)))
                            (,(1+index-of 'single-float)
-                            ,(mix-float '(%raw-instance-ref/single key i) $0f0))
+                            ,(mix-float '(%raw-instance-ref/single key i) 0f0))
                            (,(1+index-of 'double-float)
-                            ,(mix-float '(%raw-instance-ref/double key i) $0d0))
+                            ,(mix-float '(%raw-instance-ref/double key i) 0d0))
                            (,(1+index-of 'sb-kernel:complex-single-float)
                             (let ((cplx (%raw-instance-ref/complex-single key i)))
-                              ,(mix-float '(realpart cplx) $0f0)
-                              ,(mix-float '(imagpart cplx) $0f0)))
+                              ,(mix-float '(realpart cplx) 0f0)
+                              ,(mix-float '(imagpart cplx) 0f0)))
                            (,(1+index-of 'sb-kernel:complex-double-float)
                             (let ((cplx (%raw-instance-ref/complex-double key i)))
-                              ,(mix-float '(realpart cplx) $0d0)
-                              ,(mix-float '(imagpart cplx) $0d0)))))))
+                              ,(mix-float '(realpart cplx) 0d0)
+                              ,(mix-float '(imagpart cplx) 0d0)))))))
            (let* ((layout (%instance-layout key))
                   (result (layout-clos-hash layout)))
              (declare (type fixnum result))
@@ -518,7 +575,7 @@
                    ;; This is a purposely not very strong hash so that it does not make any
                    ;; distinctions that EQUALP does not make. Computing a hash of the k/v pair
                    ;; vector would incorrectly take insertion order into account.
-                   (mix (mix 103924836 (hash-table-count key))
+                   (mix (mix 103924836 (hash-table-%count key))
                         (sxhash (hash-table-test key))))
                   ((pathnamep key) (pathname-sxhash key))
                   (t
@@ -566,3 +623,8 @@
 
 ;;; Not needed post-build
 (clear-info :function :inlining-data '%sxhash-simple-substring)
+(clear-info :function :inlining-data '%sxhash-simple-substring/truncating)
+
+;; interpreter stubs that are needed as soon as possible in warm build
+(defun get-lisp-obj-address (x) (get-lisp-obj-address x))
+(defun descriptor-hash32 (x) (descriptor-hash32 x))
