@@ -42,7 +42,7 @@
 ;;;; Lisp types used by syscalls
 
 (deftype unix-pathname () 'simple-string)
-(deftype unix-fd () `(integer 0 ,most-positive-fixnum))
+(deftype unix-fd () '(unsigned-byte #-win32 31 #+win32 #.sb-vm:n-word-bits))
 
 (deftype unix-file-mode () '(unsigned-byte 32))
 (deftype unix-pid () '(unsigned-byte 32))
@@ -61,30 +61,34 @@
     ;; called directly is listed explicitly, because there are also others
     ;; that might want to be wrapped even if they don't need to be,
     ;; like sb_opendir and sb_closedir. Why are those wrapped in fact?
-    #+netbsd x
-    #-netbsd (if (member x '("sb_getrusage" ; syscall*
-                             "sb_gettimeofday" ;syscall*
-                             "sb_select" ; int-syscall
-                             "sb_getitimer" ; syscall*
-                             "sb_setitimer" ; syscall*
-                             "sb_clock_gettime" ; alien-funcall
-                             "sb_utimes") ; posix
-                         :test #'string=)
-                 (subseq x 3)
-                 x)))
+    #+(or netbsd (not 64-bit)) x
+    #-(or netbsd (not 64-bit))
+    (if (member x '("sb_getrusage"      ; syscall*
+                    "sb_gettimeofday"   ;syscall*
+                    "sb_clock_gettime"  ; alien-funcall
+                    "sb_select"         ; int-syscall
+                    "sb_getitimer"      ; syscall*
+                    "sb_setitimer"      ; syscall*
+                    "sb_utimes")        ; posix
+                :test #'string=)
+        (subseq x 3)
+        x)))
 
-(defmacro syscall ((name &rest arg-types) success-form &rest args)
+(defmacro syscall-type ((name return-type &rest arg-types) success-form &rest args)
   (when (eql 3 (mismatch "[_]" name))
     (setf name
           (concatenate 'string #+win32 "_" (subseq name 3))))
   `(locally
-    (declare (optimize (sb-c::float-accuracy 0)))
-    (let ((result (alien-funcall (extern-alien ,(libc-name-for name)
-                                               (function int ,@arg-types))
-                                ,@args)))
-      (if (minusp result)
-          (values nil (get-errno))
-          ,success-form))))
+       (declare (optimize (sb-c::float-accuracy 0)))
+     (let ((result (alien-funcall (extern-alien ,(libc-name-for name)
+                                                (function ,return-type ,@arg-types))
+                                  ,@args)))
+       (if (minusp result)
+           (values nil (get-errno))
+           ,success-form))))
+
+(defmacro syscall ((name &rest arg-types) success-form &rest args)
+  `(syscall-type (,name int ,@arg-types) ,success-form ,@args))
 
 ;;; This is like SYSCALL, but if it fails, signal an error instead of
 ;;; returning error codes. Should only be used for syscalls that will
@@ -102,6 +106,9 @@
 (defmacro int-syscall ((name &rest arg-types) &rest args)
   `(syscall (,(libc-name-for name) ,@arg-types) (values result 0) ,@args))
 
+(defmacro type-syscall ((name return-type &rest arg-types) &rest args)
+  `(syscall-type (,(libc-name-for name) ,return-type ,@arg-types) (values result 0) ,@args))
+
 (defmacro with-restarted-syscall ((&optional (value (gensym))
                                              (errno (gensym)))
                                   syscall-form &rest body)
@@ -193,8 +200,8 @@ corresponds to NAME, or NIL if there is none."
 ;;; associated with it.
 (/show0 "unix.lisp 391")
 (defun unix-close (fd)
+  (declare (type unix-fd fd))
   #+win32 (sb-win32:unixlike-close fd)
-  #-win32 (declare (type unix-fd fd))
   #-win32 (void-syscall ("close" int) fd))
 
 ;;;; stdlib.h
@@ -321,9 +328,13 @@ corresponds to NAME, or NIL if there is none."
 (defun unix-read (fd buf len)
   (declare (type unix-fd fd)
            (type index len))
-  (int-syscall (#-win32 "read" #+win32 "win32_unix_read"
-                ssize-t (* char) size-t)
-               fd buf len))
+  (type-syscall (#-win32 "read" #+win32 "win32_unix_read"
+                 ssize-t
+                 int (* char) size-t)
+                fd buf
+                (min len
+                     #+(or darwin freebsd)
+                     (1- (expt 2 31)))))
 
 ;;; UNIX-WRITE accepts a file descriptor, a buffer, an offset, and the
 ;;; length to write. It attempts to write len bytes to the device
@@ -338,12 +349,14 @@ corresponds to NAME, or NIL if there is none."
            (type index offset len))
   (flet ((%write (sap)
            (declare (system-area-pointer sap))
-           (int-syscall (#-win32 "write" #+win32 "win32_unix_write"
-                         ssize-t (* char) size-t)
-                        fd
-                        (with-alien ((ptr (* char) sap))
-                          (addr (deref ptr offset)))
-                        len)))
+           (type-syscall (#-win32 "write" #+win32 "win32_unix_write"
+                          ssize-t int (* char) size-t)
+                         fd
+                         (with-alien ((ptr (* char) sap))
+                           (addr (deref ptr offset)))
+                         (min len
+                              #+(or darwin freebsd)
+                              (1- (expt 2 31))))))
     (etypecase buf
       ((simple-array * (*))
        (with-pinned-objects (buf)
@@ -634,15 +647,21 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
       (note-dangerous-wait "poll(2)"))
     (let ((events (ecase direction
                     (:input (logior pollin pollpri))
-                    (:output pollout))))
+                    (:output pollout)))
+          (deadline (if (minusp to-msec)
+                        to-msec
+                        (+ (* (get-universal-time) 1000) to-msec))))
       (with-alien ((fds (struct pollfd)))
         (with-restarted-syscall (count errno)
-          (progn
+          (let ((timeout (if (minusp to-msec)
+                             -1
+                             (max 0 (- deadline (* 1000 (get-universal-time)))))))
+            (declare (fixnum timeout))
             (setf (slot fds 'fd) fd
                   (slot fds 'events) events
                   (slot fds 'revents) 0)
             (int-syscall ("poll" (* (struct pollfd)) int int)
-                         (addr fds) 1 to-msec))
+                         (addr fds) 1 timeout))
           (if (zerop errno)
               (let ((revents (slot fds 'revents)))
                 (or (and (eql 1 count) (logtest events revents))
@@ -714,29 +733,34 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
 
 #-os-provides-poll
 (defun unix-simple-poll (fd direction to-msec)
-  (multiple-value-bind (to-sec to-usec)
-      (if (minusp to-msec)
-          (values nil nil)
-          (multiple-value-bind (to-sec to-msec2) (truncate to-msec 1000)
-            (values to-sec (* to-msec2 1000))))
-    (with-restarted-syscall (count errno)
-      (with-alien ((fds (struct fd-set)))
-        (fd-zero fds)
-        (fd-set fd fds)
-        (multiple-value-bind (read-fds write-fds)
-            (ecase direction
-              (:input
-               (values (addr fds) nil))
-              (:output
-               (values nil (addr fds))))
-          (unix-fast-select (1+ fd)
-                                    read-fds write-fds nil
-                                    to-sec to-usec)))
-      (case count
-        ((1) t)
-        ((0) nil)
-        (otherwise
-         (error "Syscall select(2) failed on fd ~D: ~A" fd (strerror)))))))
+  (flet ((msec-to-sec-usec (msec)
+           (multiple-value-bind (sec msec2) (truncate msec 1000)
+             (values sec (* msec2 1000)))))
+    (let ((deadline (if (minusp to-msec)
+                        to-msec
+                        (+ (* (get-universal-time) 1000) to-msec))))
+      (with-restarted-syscall (count errno)
+        (with-alien ((fds (struct fd-set)))
+          (fd-zero fds)
+          (fd-set fd fds)
+          (multiple-value-bind (to-sec to-usec)
+              (if (minusp to-msec)
+                  (values nil nil)
+                  (msec-to-sec-usec (max 0 (- deadline (* 1000 (get-universal-time))))))
+            (multiple-value-bind (read-fds write-fds)
+                (ecase direction
+                  (:input
+                   (values (addr fds) nil))
+                  (:output
+                   (values nil (addr fds))))
+              (unix-fast-select (1+ fd)
+                                read-fds write-fds nil
+                                to-sec to-usec))))
+        (case count
+          ((1) t)
+          ((0) nil)
+          (otherwise
+           (error "Syscall select(2) failed on fd ~D: ~A" fd (strerror))))))))
 
 ;;;; sys/stat.h
 
@@ -824,7 +848,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
              (%extract-stat-results (addr buf))
              name (addr buf))))
 (defun unix-fstat (fd)
-  #-win32
   (declare (type unix-fd fd))
   (#-win32 funcall #+win32 sb-win32::call-with-crt-fd
    (lambda (fd)
@@ -893,15 +916,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
                  seconds)
   nil)
 
-;;;; sys/time.h
-
-;;; Structure crudely representing a timezone. KLUDGE: This is
-;;; obsolete and should never be used.
-(define-alien-type nil
-  (struct timezone
-    (tz-minuteswest int)                ; minutes west of Greenwich
-    (tz-dsttime int)))                  ; type of dst correction
-
 
 ;; Type of the second argument to `getitimer' and
 ;; the second and third arguments `setitimer'.
@@ -910,10 +924,6 @@ avoiding atexit(3) hooks, etc. Otherwise exit(2) is called."
     (it-interval (struct timeval))      ; timer interval
     (it-value (struct timeval))))       ; current value
 
-(defconstant itimer-real 0)
-(defconstant itimer-virtual 1)
-(defconstant itimer-prof 2)
-
 #-win32
 (defun unix-getitimer (which)
   "UNIX-GETITIMER returns the INTERVAL and VALUE slots of one of
@@ -1026,9 +1036,9 @@ the UNIX epoch (January 1st 1970.)"
           ;; offers, and COARSE is about twice as fast, so use that, but only for linux.
           ;; BSD has something similar.
           #-avoid-clock-gettime
-          (clock-gettime #+linux clock-monotonic-coarse #-linux clock-monotonic)
-          #+avoid-clock-gettime
-          (multiple-value-bind (c-sec c-usec) (get-time-of-day) (values c-sec (* c-usec 1000)))
+        (clock-gettime #+linux clock-monotonic-coarse #-linux clock-monotonic)
+        #+avoid-clock-gettime
+        (multiple-value-bind (c-sec c-usec) (get-time-of-day) (values c-sec (* c-usec 1000)))
 
         #+64-bit ;; I know that my math is valid for 64-bit.
         (declare (optimize (sb-c::type-check 0)))
@@ -1039,21 +1049,21 @@ the UNIX epoch (January 1st 1970.)"
                (+ (the fixnum (* delta-sec internal-time-units-per-second))
                   (truncate delta-nsec nanoseconds-per-internal-time-unit))))
 
-  ;; There are two optimizations here that actually matter on 32-bit systems:
-  ;;  (1) subtract the epoch from seconds and milliseconds separately,
-  ;;  (2) avoid consing a new bignum if the result is unchanged.
-  ;;
-  ;; Thanks to James Anderson for the optimization hint.
-  ;;
-  ;; Yes, it is possible to a computation to be GET-INTERNAL-REAL-TIME
-  ;; bound.
-  ;;
-  ;; --NS 2007-04-05
+        ;; There are two optimizations here that actually matter on 32-bit systems:
+        ;;  (1) subtract the epoch from seconds and milliseconds separately,
+        ;;  (2) avoid consing a new bignum if the result is unchanged.
+        ;;
+        ;; Thanks to James Anderson for the optimization hint.
+        ;;
+        ;; Yes, it is possible to a computation to be GET-INTERNAL-REAL-TIME
+        ;; bound.
+        ;;
+        ;; --NS 2007-04-05
         #-64-bit
         (symbol-macrolet ((observed-sec
-                           (sb-thread::thread-observed-internal-real-time-delta-sec thr))
+                            (sb-thread::thread-observed-internal-real-time-delta-sec thr))
                           (observed-msec
-                           (sb-thread::thread-observed-internal-real-time-delta-millisec thr))
+                            (sb-thread::thread-observed-internal-real-time-delta-millisec thr))
                           (time (sb-thread::thread-internal-real-time thr)))
           (let* ((delta-sec (- c-sec (slot base 'tv-sec)))
                  ;; I inadvertently had too many THE casts in here, so I'd prefer
