@@ -137,16 +137,6 @@
 ;;; when it reads #- and #+ syntax)
 (declaim (type function *in-target-compilation-mode-fn*))
 (defvar *in-target-compilation-mode-fn*)
-
-;;; a function with the same calling convention as CL:COMPILE-FILE, to be
-;;; used to translate ordinary Lisp source files into target object files
-(declaim (type function *target-compile-file*))
-(defvar *target-compile-file*)
-
-;;; designator for a function with the same calling convention as
-;;; SB-C:ASSEMBLE-FILE, to be used to translate assembly files into target
-;;; object files
-(defvar *target-assemble-file*)
 
 ;;;; some tools
 
@@ -319,12 +309,6 @@
           (pushnew :immobile-code sb-xc:*features*))
         (when (target-featurep :64-bit)
           (push :compact-symbol sb-xc:*features*))
-        (when (target-featurep :64-bit)
-          ;; Considering that a single config file governs rv32 and rv64, I don't
-          ;; know how to make this properly configurable. In theory, 32-bit builds could
-          ;; have a salted hash (gaining 3 bits by making the hash slot raw), but
-          ;; they don't, so in light of things, this is a valid criterion.
-          (push :salted-symbol-hash sb-xc:*features*))
         (when (target-featurep '(:and :sb-thread (:or (:and :darwin (:not (:or :ppc :x86))) :openbsd)))
           (push :os-thread-stack sb-xc:*features*))
         (when (target-featurep '(:and :x86 :int4-breakpoints))
@@ -353,6 +337,8 @@
           (push :round-float sb-xc:*features*))
         (when (target-featurep '(:and :arm64 :darwin))
           (push :arm-v8.1 backend-subfeatures))
+        (when (target-featurep '(:and :ppc64 :little-endian))
+          (push :fsqrt backend-subfeatures))
 
         ;; Putting arch and gc choice first is visually convenient, versus
         ;; having to parse a random place in the line to figure out the value
@@ -381,6 +367,7 @@
           ":SB-THREAD not supported on selected architecture")
          ("(and mark-region-gc (not (or x86-64 arm64)))"
           "mark-region is not supported on selected architecture")
+         ("(and mark-region-gc (not sb-thread))" "mark-region requires threads")
          ("(and (not sb-thread) (or arm64 ppc64))"
           "The selected architecture requires :SB-THREAD")
          ("(and gencgc cheneygc)"
@@ -398,6 +385,8 @@
           ;; It sorta kinda works to have both, but there should be no need,
           ;; and it's not really supported.
           "At most one interpreter can be selected")
+         ("(and immobile-space permgen)"
+          ":IMMOBILE-SPACE and :PERMGEN are mutually exclusive options")
          ("(and compact-instance-header (not (or permgen immobile-space)))"
           ":COMPACT-INSTANCE-HEADER requires :IMMOBILE-SPACE feature")
          ("(and immobile-code (not immobile-space))"
@@ -408,6 +397,10 @@
           ":SYSTEM-TLABS requires SB-THREAD")
          ("(and sb-futex (not sb-thread))"
           "Can't enable SB-FUTEX on platforms lacking thread support")
+         ("(and relocatable-static-space (not (or arm64 x86-64)))"
+          "Relocatable-static-space not supported for chosen architecture")
+         ("(and relocatable-static-space arm64 (not immobile-space))"
+          "Relocatable-static-space requires immobile-space")
          ;; There is still hope to make multithreading on DragonFly x86-64
          ("(and sb-thread x86 dragonfly)"
           ":SB-THREAD not supported on selected architecture")))
@@ -626,12 +619,12 @@
                                                  (muffle-warning c)))))
                               (apply #'compile-file args)))
                           #-(or abcl ccl) #'compile-file)
-                         (:target-compile (if (find :assem flags)
-                                              *target-assemble-file*
-                                              *target-compile-file*))))
+                         (:target-compile
+                          (intern (if (find :assem flags) "ASSEMBLE-FILE" "COMPILE-FILE")
+                                  "SB-C"))))
          (trace-file (if (find :trace-file flags) t nil))
          (block-compile (if (find :block-compile flags) t :specified)))
-    (declare (type function compilation-fn))
+    (declare (type (or function symbol) compilation-fn))
 
     (ensure-directories-exist obj :verbose cl:*compile-print*) ; host's value
 
@@ -818,7 +811,7 @@
 (defun target-compile-file (filename)
   (funcall *in-target-compilation-mode-fn*
            (lambda ()
-             (funcall *target-compile-file* filename))))
+             (funcall (intern "COMPILE-FILE" "SB-XC") filename))))
 (compile 'target-compile-file)
 
 (defvar *math-ops-memoization* (make-hash-table :test 'equal))
@@ -892,10 +885,10 @@
   #+unix "tools-for-build/perfecthash"
   #+win32 "tools-for-build/perfecthash.exe")
 
-#+sbcl (when (and (probe-file (perfect-hash-generator-program))
-                  (find-symbol "RUN-PROGRAM" "SB-EXT"))
-         (pushnew :use-host-hash-generator cl:*features*)
-         (setq *perfect-hash-generator-mode* :RECORD))
+#+(or sbcl ecl ccl clisp cmucl)
+(when (probe-file (perfect-hash-generator-program))
+  (pushnew :use-host-hash-generator cl:*features*)
+  (setq *perfect-hash-generator-mode* :RECORD))
 
 ;;; I want this to work using the host-native readtable if sb-cold:*xc-readtable*
 ;;; isn't established. The caller should bind *READTABLE* to ours if reading
@@ -937,58 +930,121 @@
           (error "hash generator duplicates: ~D" errors))))))
 (compile 'preload-perfect-hash-generator)
 
+#+use-host-hash-generator
+(defun run-perfecthash (input)
+  (with-output-to-string (result)
+    (flet (#+sbcl
+           (launch ()
+             (let ((process (sb-ext:run-program (perfect-hash-generator-program)
+                                                '()
+                                                :input :stream :output :stream
+                                                :wait nil
+                                                :allow-other-keys t
+                                                :use-posix-spawn t)))
+               (values (sb-ext:process-output process)
+                       (sb-ext:process-input process)
+                       process)))
+           #+sbcl
+           (wait (process)
+             (sb-ext:process-wait process)
+             (sb-ext:process-close process)
+             (unless (zerop (sb-ext:process-exit-code process))
+               (error "Error running perfecthash: exit code ~D"
+                      (sb-ext:process-exit-code process))))
+           #+cmu
+           (launch ()
+             (let ((process (ext:run-program (perfect-hash-generator-program)
+                                             '()
+                                             :input :stream :output :stream
+                                             :wait nil)))
+               (values (ext:process-output process)
+                       (ext:process-input process)
+                       process)))
+           #+cmu
+           (wait (process)
+             (ext:process-wait process)
+             (ext:process-close process)
+             (unless (zerop (ext:process-exit-code process))
+               (error "Error running perfecthash: exit code ~D"
+                      (ext:process-exit-code process))))
+           #+clisp
+           (launch ()
+             (multiple-value-bind (io-stream input-stream output-stream)
+                 (ext:run-program (perfect-hash-generator-program)
+                                  :input :stream :output :stream)
+               (declare (ignore io-stream))
+               (values input-stream output-stream nil)))
+           #+(or clisp ccl)
+           (wait (process)
+             process)
+           #+ecl
+           (launch ()
+             (let ((process
+                     (nth-value 2 (ext:run-program (perfect-hash-generator-program)
+                                                   ()
+                                                   :input :stream :output :stream
+                                                   :wait nil))))
+               (values (ext:external-process-output process)
+                       (ext:external-process-input process)
+                       process)))
+           #+ecl
+           (wait (process)
+             (multiple-value-bind (status code) (ext:external-process-wait process t)
+               (unless (and (eq status :exited)
+                            (zerop code))
+                 (error "Error running perfecthash: exit code ~D" code))))
+           #+ccl
+           (launch ()
+             (let ((process
+                     (ccl:run-program (perfect-hash-generator-program)
+                                      ()
+                                      :wait nil
+                                      :input :stream :output :stream)))
+               (values (ccl:external-process-output-stream process)
+                       (ccl:external-process-input-stream process)
+                       process))))
+      (multiple-value-bind (input-stream output-stream process) (launch)
+        (format output-stream "~{~X~%~}" (coerce input 'list))
+        (close output-stream)
+        (loop for char = (read-char input-stream nil)
+              while char
+              do (write-char char result))
+        (close input-stream)
+        (wait process)))))
+
+
 (defun emulate-generate-perfect-hash-sexpr (array identifier digest)
   (declare #-use-host-hash-generator (ignore identifier))
-  (let (computed)
-    (declare (ignorable computed))
-    ;; Entries are written to disk with hashes sorted in ascending order so that
-    ;; comparing as sets can be done using EQUALP.
-    ;; Sort nondestructively in case something else looks at the value as supplied.
-    (let* ((canonical-array (sort (copy-seq array) #'<))
-           (match (assoc (cons digest canonical-array) *perfect-hash-generator-memo*
-                         :test #'equalp)))
-      (when match
-        (return-from emulate-generate-perfect-hash-sexpr (cddr match)))
-      (ecase *perfect-hash-generator-mode*
-        (:playback
-         (error "perfect hash file is missing a needed entry for ~x" array))
-        (:record
-         ;; This will only display anything when we didn't have the data,
-         ;; so it's actually not too "noisy" in a normal build.
-         #+use-host-hash-generator
-         (let ((output (make-string-output-stream))
-               (process
-                (sb-ext:run-program (perfect-hash-generator-program)
-                                    '("perfecthash")
-                                    ;; win32 misbehaves with :input string-stream
-                                    :input :stream :output :stream
-                                    :wait nil
-                                    :allow-other-keys t
-                                    :use-posix-spawn t)))
-           (format (sb-ext:process-input process) "~{~X~%~}" (coerce array 'list))
-           (close (sb-ext:process-input process))
-           (loop for char = (read-char (sb-ext:process-output process) nil)
-                 while char
-                 do (write-char char output))
-           (sb-ext:process-wait process)
-           (sb-ext:process-close process)
-           (unless (zerop (sb-ext:process-exit-code process))
-             (error "Error running perfecthash: exit code ~D"
-                    (sb-ext:process-exit-code process)))
-           (let* ((string (get-output-stream-string output))
-                  ;; don't need the final newline, it looks un-lispy in the file
-                  (l (length string)))
-             (assert (char= (char string (1- l)) #\newline))
-             (setq computed (subseq string 0 (1- l))))
-           (let ((*print-right-margin* 200) (*print-level* nil) (*print-length* nil))
-             (format t "~&Recording perfect hash:~%~S~%~X~%"
-                     identifier array))
-           (setf *perfect-hash-generator-memo*
-                 (nconc *perfect-hash-generator-memo*
-                        (list (list* (cons digest canonical-array)
-                                     identifier
-                                     computed))))
-           computed))))))
+  ;; Entries are written to disk with hashes sorted in ascending order so that
+  ;; comparing as sets can be done using EQUALP.
+  ;; Sort nondestructively in case something else looks at the value as supplied.
+  (let* ((canonical-array (sort (copy-seq array) #'<))
+         (match (assoc (cons digest canonical-array) *perfect-hash-generator-memo*
+                       :test #'equalp)))
+    (when match
+      (return-from emulate-generate-perfect-hash-sexpr (cddr match)))
+    (ecase *perfect-hash-generator-mode*
+      (:playback
+       (error "perfect hash file is missing a needed entry for ~x" array))
+      #+use-host-hash-generator
+      (:record
+       ;; This will only display anything when we didn't have the data,
+       ;; so it's actually not too "noisy" in a normal build.
+       (let ((string (run-perfecthash array))
+             computed)
+         (let* (;; don't need the final newline, it looks un-lispy in the file
+                (l (length string)))
+           (assert (char= (char string (1- l)) #\newline))
+           (setq computed (subseq string 0 (1- l))))
+         (let ((*print-right-margin* 200) (*print-level* nil) (*print-length* nil))
+           (format t "~&Recording perfect hash:~%~S~%~X~%"
+                   identifier array))
+         (setf *perfect-hash-generator-memo*
+               (nconc *perfect-hash-generator-memo*
+                      (list (list* (cons digest canonical-array)
+                                   identifier
+                                   computed))))
+         computed)))))
 
 ;;; Unlike xfloat-math which expresses universal truths, the perfect-hash file
 ;;; expresses facts about the behavior of a _particular_ SBCL revision.
@@ -1081,7 +1137,7 @@
       (let ((array (cdar entry)))
         (assert (not (gethash array uniqueness-checker)))
         (setf (gethash array uniqueness-checker) t))))
-  #+use-host-hash-generator
+  #+(and use-host-hash-generator sbcl)
   (when (eq *perfect-hash-generator-mode* :record)
     (save-perfect-hashfuns (perfect-hash-generator-journal :output)
                            *perfect-hash-generator-memo*))
