@@ -857,6 +857,13 @@
        (eq (numeric-type-class x) 'integer)
        (eq (numeric-type-complexp x) :real)))
 
+(defun ratio-type-p (x)
+  (declare (type ctype x))
+  (and (numeric-type-p x)
+       (eq (numeric-type-class x) 'rational)
+       (eq (numeric-type-complexp x) :real)
+       (csubtypep x (specifier-type 'ratio))))
+
 ;;; Given that an inequality holds on values of type X and Y, return a
 ;;; new type for X. If GREATER is true, then X was greater than Y,
 ;;; otherwise less. If OR-EQUAL is true, then the inequality was
@@ -1062,6 +1069,8 @@
                       (when (and ref
                                  (constant-p other))
                         (change-ref-leaf ref other)
+                        (when (eq (node-derived-type ref) *empty-type*)
+                          (pushnew ref *blocks-to-terminate*))
                         (return-from type-from-constraints))
                       (intersect-result other-type))
                      ((constant-p other)
@@ -1268,12 +1277,13 @@
                             (conset-add-constraint gen 'typep var type nil)))
                         (maybe-add-eql-var-var-constraint var val gen)
                         (add-var-result-constraints var val gen)))
-           ((nil)
+           ((nil optional)
             (loop for var in (lambda-vars fun)
                   for type = (leaf-defined-type var)
                   do
                   (when (and (lambda-var-constraints var)
-                             (type-for-constraints-p type))
+                             (type-for-constraints-p type)
+                             (not (lambda-var-arg-info var)))
                     (conset-add-constraint gen 'typep var type nil))))
            (mv-let
             (add-mv-let-result-constraints (lvar-dest (node-lvar (first (lambda-refs fun)))) fun gen)))))
