@@ -621,7 +621,10 @@
                      (template-more-results-type info) "results")
       (check-tn-refs (vop-temps vop) vop t 0 t "temps")
       (unless (or (= (length (vop-codegen-info vop))
-                     (template-info-arg-count info))
+                     (+ (template-info-arg-count info)
+                        (if (vop-info-gc-barrier info)
+                            1
+                            0)))
                   ;; Allow these 2 allocator vops to take an undeclared info arg
                   (member (vop-info-name info) '(sb-vm::fixed-alloc sb-vm::var-alloc))
                   ;; FIXME: The current representation for conditional
@@ -1095,7 +1098,12 @@
                    (write-string "}"))))
              (write-char #\space)))
           (cdynamic-extent
-           (format t "dynamic extent ~S" (dynamic-extent-values node))))
+           (format t "dynamic extent ~S" (dynamic-extent-values node))
+           (let ((info (dynamic-extent-info node)))
+             (write-string " {info: ")
+             (when info
+               (print-lvar info))
+             (write-string "}"))))
         (when (and *debug-print-types*
                    (valued-node-p node))
           (write-char #\space)
@@ -1113,25 +1121,42 @@
       (format t "cleanup ~s~%" (cleanup-kind (block-end-cleanup block)))))
   (values))
 
+(defun tn-write-count (tn)
+  (loop for ref = (tn-writes tn) then (tn-ref-next ref)
+        while ref
+        count t))
+
+(defun tn-read-count (tn)
+  (loop for ref = (tn-reads tn) then (tn-ref-next ref)
+        while ref
+        count t))
+
 ;;; Print the guts of a TN. (logic shared between PRINT-OBJECT (TN T)
 ;;; and printers for compound objects which contain TNs)
 (defun print-tn-guts (tn &optional (stream *standard-output*))
   (declare (type tn tn))
-  (let ((leaf (tn-leaf tn)))
-    (cond (leaf
-           (print-leaf leaf stream)
-           (format stream "!~D" (tn-id tn)))
-          (t
-           (format stream "t~D" (tn-id tn))))
-    (when (and (tn-sc tn) (tn-offset tn))
-      (format stream "[~A]" (location-print-name tn)))
-    (format stream " ~s~@[ ~a~]~@[ ~a~]" (tn-kind tn)
-            (and *debug-print-types*
-                 (and (tn-sc tn)
-                      (sc-name (tn-sc tn))))
-            (and *debug-print-types*
-                 (and (tn-primitive-type tn)
-                      (primitive-type-name (tn-primitive-type tn)))))))
+  (flet ((print-tn-name (tn)
+           (let ((leaf (tn-leaf tn)))
+             (cond (leaf
+                    (print-leaf leaf stream)
+                    (format stream "!~D" (tn-id tn)))
+                   (t
+                    (format stream "t~D" (tn-id tn))))
+             (when (and (tn-sc tn) (tn-offset tn))
+               (format stream "[~A]" (location-print-name tn))))))
+    (print-tn-name tn)
+    (format stream " ~s" (tn-kind tn))
+    (when (eq (tn-kind tn) :alias)
+      (format stream " to ")
+      (print-tn-name (tn-save-tn tn)))
+    (when *debug-print-types*
+      (format stream "~@[ ~a~]~@[ ~a~]"
+              (and (tn-sc tn)
+                   (sc-name (tn-sc tn)))
+              (and (tn-primitive-type tn)
+                   (primitive-type-name (tn-primitive-type tn)))))
+    #+nil
+    (format stream "/r~aw~a/" (tn-read-count tn) (tn-write-count tn))))
 
 ;;; Print the TN-REFs representing some operands to a VOP, linked by
 ;;; TN-REF-ACROSS.
@@ -1482,6 +1507,14 @@ is replaced with replacement."
   (do-conset-elements (con conset)
     (print-constraint con kind)))
 
+(defun print-conset-difference (set1 set2)
+  (let ((diff1 (conset-difference (copy-conset set1) set2))
+        (diff2 (conset-difference (copy-conset set2) set1)))
+    (format t "Not in set1~%")
+    (print-conset diff1)
+    (format t "Not in set2~%")
+    (print-conset diff2)))
+
 (defun print-constraints (component &optional kind)
   (do-blocks (block component)
     (handler-case (progn
