@@ -96,6 +96,9 @@
 
 ;;;; miscellaneous "sub-primitives"
 
+(defknown descriptor-hash32 (t) #+64-bit (unsigned-byte 32) #-64-bit (unsigned-byte 29)
+          (flushable always-translatable))
+
 (defknown %sp-string-compare
   (simple-string simple-string index (or null index) index (or null index))
   (values index fixnum)
@@ -123,9 +126,6 @@
 ;;; not only symbols. The value is reliable only if the object is a symbol.
 (defknown hash-as-if-symbol-name (t) symbol-name-hash (flushable movable always-translatable))
 
-(defknown %set-symbol-hash (symbol hash-code)
-  t ())
-
 ;;; SYMBOL-PACKAGE-ID for #+compact-symbol demands a vop which avoids loading
 ;;; a raw bit value in a descriptor register (the SLOT vop returns a descriptor)
 (defknown symbol-package-id (symbol) (unsigned-byte 16))
@@ -147,7 +147,7 @@
 
 ;;; Return the length of VECTOR.
 ;;; Ordinary code should prefer to use (LENGTH (THE VECTOR FOO)) instead.
-(defknown vector-length (vector) index (flushable dx-safe))
+(defknown vector-length ((read-only vector)) index (flushable dx-safe))
 
 (defknown vector-sap ((simple-unboxed-array (*))) system-area-pointer
   (flushable))
@@ -195,11 +195,15 @@
 (defknown (%array-rank= widetag=) (t t) boolean
   (flushable))
 
+(defknown vector-data (array index) (values simple-array index)
+  (flushable))
+
 (defknown simple-array-header-of-rank-p (t %array-rank) boolean
   (flushable))
 (defknown sb-kernel::check-array-shape (simple-array list)
   (simple-array)
   (flushable no-verify-arg-count)
+  :derive-type #'result-type-first-arg
   :result-arg 0)
 
 (defknown (%make-instance %make-instance/mixed) (index) instance
@@ -232,6 +236,10 @@
   (flushable always-translatable))
 (defknown (%instance-ref-eq) (instance index t) boolean
   (flushable always-translatable))
+;; This predicates sounds as though the argument restriction would be INSTANCE,
+;; but it's lenient because it can perform a lowtag test on one (but not both) args.
+(defknown (%instance-types=) (t t) boolean
+  (flushable always-translatable))
 (defknown %instance-set (instance index t) (values) (always-translatable))
 (defknown update-object-layout (t) layout)
 
@@ -287,16 +295,18 @@
                            word index
                            ;; The number of words is later converted
                            ;; to bytes, make sure it fits.
-                           (and index
-                                (mod #.(- (expt 2
-                                                (- sb-vm:n-word-bits
-                                                   sb-vm:word-shift
-                                                   ;; all the allocation routines expect a signed word
-                                                   1))
-                                          ;; The size is double-word aligned, which is done by adding
-                                          ;; (1- (/ sb-vm:n-word-bits 2)) and then masking.
-                                          ;; Make sure addition doesn't overflow.
-                                          3))))
+                           #.(if (fixnump (ash array-dimension-limit 7))
+                                 `(and unsigned-byte fixnum)
+                                 `(and index
+                                       (mod ,(- (expt 2
+                                                      (- sb-vm:n-word-bits
+                                                         sb-vm:word-shift
+                                                         ;; all the allocation routines expect a signed word
+                                                         1))
+                                                ;; The size is double-word aligned, which is done by adding
+                                                ;; (1- (/ sb-vm:n-word-bits 2)) and then masking.
+                                                ;; Make sure addition doesn't overflow.
+                                                3)))))
     (simple-array * (*))
     (flushable movable))
 
@@ -711,9 +721,29 @@
   sb-interpreter::interpreted-fun-prototype (flushable))
 
 
-(defknown %data-vector-and-index (array index)
-                                 (values (simple-array * (*)) index)
-                                 (foldable flushable))
+(defknown %data-vector-and-index
+    (array index)
+    (values (simple-array * (*)) index)
+    (foldable flushable))
+
+(defknown %data-vector-and-index/check-bound (array index)
+    (values (simple-array * (*)) index)
+    (foldable))
+
+(defknown %data-vector-and-index/known
+    (array index)
+    (values array index)
+    (flushable always-translatable))
+
+;;; Checks for and adjusts fill-pointer for vector-pop/push and
+;;; returns the underlying simple data vector.
+(defknown %data-vector-pop
+    (array)
+    (values (simple-array * (*)) index))
+
+(defknown %data-vector-push
+    (array)
+    (values (simple-array * (*)) (or null index)))
 
 (defknown restart-point (t) t ())
 
@@ -735,10 +765,10 @@
   (movable foldable unboxed-return))
 
 (defknown make-single-float ((signed-byte 32)) single-float
-  (movable flushable))
+  (movable flushable foldable))
 
 (defknown make-double-float ((signed-byte 32) (unsigned-byte 32)) double-float
-  (movable flushable))
+  (movable flushable foldable))
 
 (defknown single-float-bits (single-float) (signed-byte 32)
   (movable foldable flushable))
@@ -746,7 +776,7 @@
 #+64-bit
 (progn
 (defknown %make-double-float ((signed-byte 64)) double-float
-  (movable flushable))
+  (movable flushable foldable))
 (defknown double-float-bits (double-float) (signed-byte 64)
   (movable foldable flushable)))
 
