@@ -30,7 +30,14 @@
             (compiler-notify "~@<unable to ~2I~_~A ~I~_because: ~2I~_~?~:>"
                              note (first what) (rest what)))
            ((valid-fun-use node what
-                           :argument-test #'types-equal-or-intersect
+                           :argument-test (lambda (arg-type type)
+                                            ;; Don't bother with LIST not matching (OR NULL ...)
+                                            (unless (and (eq arg-type (specifier-type 'list))
+                                                         (neq type (specifier-type 'list))
+                                                         (and (union-type-p type)
+                                                              (find (specifier-type 'null)
+                                                                    (union-type-types type))))
+                                              (types-equal-or-intersect arg-type type)))
                            :result-test #'values-types-equal-or-intersect)
             (collect ((messages))
               (flet ((give-grief (string &rest stuff)
@@ -155,10 +162,6 @@
       (> two-arg->)
       (<= two-arg-<=)
       (>= two-arg->=)
-      #+sb-xc-host
-      ,@'((sb-xc:< two-arg-<)
-          (sb-xc:= two-arg-=)
-          (sb-xc:> two-arg->))
       (char-equal two-arg-char-equal)
       (char-greaterp two-arg-char-greaterp)
       (char-lessp two-arg-char-lessp)
@@ -231,6 +234,8 @@
                   (unlink-node if)
                   (%delete-lvar-use combination)
                   (use-lvar combination ref-lvar)
+                  (push "unwrap-predicates" (node-source-path con-ref))
+                  (push "unwrap-predicates" (node-source-path alt-ref))
                   (link-blocks block (node-block next)))))))))))
 
 ;;; Convert function designators to functions in calls to known functions
@@ -255,7 +260,7 @@
                (labels ((translate-two-args (name)
                           (and (eql arg-count 2)
                                (not (fun-lexically-notinline-p name (node-lexenv node)))
-                               (cadr (assoc name *two-arg-functions*))))
+                               (cadr (assoc (uncross name) *two-arg-functions*))))
                         (translate (ref)
                           (let* ((leaf (ref-leaf ref))
                                  (fun-name (and (constant-p leaf)
@@ -294,8 +299,9 @@
 (defun change-full-call (combination new-fun-name)
   (let ((ref (lvar-uses (combination-fun combination))))
     (when (ref-p ref)
-      (setf (combination-fun-info combination)
-            (fun-info-or-lose new-fun-name))
+      (when (combination-fun-info combination)
+        (setf (combination-fun-info combination)
+              (fun-info-or-lose new-fun-name)))
       (change-ref-leaf
        ref
        (find-free-fun new-fun-name ""))
@@ -304,37 +310,49 @@
 (defun rewrite-full-call (combination)
   (let ((combination-name (lvar-fun-name (combination-fun combination) t))
         (args (combination-args combination)))
-    (when (eq (combination-kind combination) :known)
-      (let ((two-arg (assoc combination-name *two-arg-functions*)))
-        (when (and two-arg
-                   (= (length args) 2))
-          (destructuring-bind (name two-arg &optional types typed-two-arg) two-arg
-            (declare (ignore name))
-            (when (and types
-                       (loop for arg in args
-                             for type in types
-                             always (csubtypep (lvar-type arg) type)))
-              (setf two-arg typed-two-arg))
-            (change-full-call combination two-arg))))
-      (let ((lvar (node-lvar combination)))
-        (when (or (lvar-single-value-p lvar)
-                  (mv-bind-unused-p lvar 1))
-          (let ((single-value-fun (getf '(truncate sb-kernel::truncate1
-                                          floor sb-kernel::floor1
-                                          ceiling sb-kernel::ceiling1)
-                                        combination-name)))
-            (when single-value-fun
-              (unless (cdr args)
-                (let* ((leaf (find-constant 1))
-                       (ref (make-ref leaf))
-                       (lvar (make-lvar combination)))
-                  (use-lvar ref lvar)
-                  (push ref (leaf-refs leaf))
-                  (insert-ref-before ref combination-name)
-                  (setf (cdr args) (list lvar))))
-              (change-full-call combination single-value-fun)
-              (setf (node-derived-type combination)
-                    (make-single-value-type (single-value-type (node-derived-type combination)))))))))))
+    (cond ((eq (combination-kind combination) :known)
+           (let ((two-arg (assoc (uncross combination-name) *two-arg-functions*)))
+             (when (and two-arg
+                        (= (length args) 2))
+               (destructuring-bind (name two-arg &optional types typed-two-arg) two-arg
+                 (declare (ignore name))
+                 (when (and types
+                            (loop for arg in args
+                                  for type in types
+                                  always (csubtypep (lvar-type arg) type)))
+                   (setf two-arg typed-two-arg))
+                 (change-full-call combination two-arg))))
+           (let ((lvar (node-lvar combination)))
+             (when (or (lvar-single-value-p lvar)
+                       (mv-bind-unused-p lvar 1))
+               (let ((single-value-fun (getf '(truncate sb-kernel::truncate1
+                                               floor sb-kernel::floor1
+                                               ceiling sb-kernel::ceiling1
+                                               round sb-kernel::round1
+                                               ftruncate sb-kernel::ftruncate1
+                                               ffloor sb-kernel::ffloor1
+                                               fceiling sb-kernel::fceiling1
+                                               fround sb-kernel::fround1)
+                                             combination-name)))
+                 (when single-value-fun
+                   (unless (cdr args)
+                     (setf (cdr args)
+                           (list (insert-ref-before (find-constant 1) combination))))
+                   (change-full-call combination single-value-fun)
+                   (setf (node-derived-type combination)
+                         (make-single-value-type (single-value-type (node-derived-type combination)))))))))
+          ((and (eq (combination-kind combination) :full)
+                (not (fun-lexically-notinline-p combination-name (node-lexenv combination))))
+           (let ((specialized (or (info :function :specialized-xep combination-name)
+                                  (let ((specialized (assoc 'sb-impl::specialized-xep
+                                                            (lexenv-user-data (node-lexenv combination)))))
+                                    (when (eq (cadr specialized) combination-name)
+                                      (cddr specialized))))))
+
+             (when (and specialized
+                        (= (length args)
+                           (length (first specialized))))
+               (change-full-call combination `(sb-impl::specialized-xep ,combination-name ,@specialized))))))))
 
 ;;; The %other-pointer-subtype-p optimizer in ir2opt combines multiple
 ;;; checks for other-pointer into a single widetag load.
