@@ -607,11 +607,17 @@
            while ,node-var
            do (progn ,@body))))
 
-(defmacro do-nested-cleanups ((cleanup-var block &optional return-value)
+;;; Walk up the cleanup nesting starting from the cleanup in effect at
+;;; BLOCK-OR-NODE, binding CLEANUP-VAR to each cleanup.
+(defmacro do-nested-cleanups ((cleanup-var block-or-node &optional return-value)
                               &body body)
-  `(block nil
-     (map-nested-cleanups
-      (lambda (,cleanup-var) ,@body) ,block ,return-value)))
+  (once-only ((block-or-node block-or-node))
+    `(do ((,cleanup-var (if (node-p ,block-or-node)
+                            (node-enclosing-cleanup ,block-or-node)
+                            (block-end-cleanup ,block-or-node))
+                        (node-enclosing-cleanup (cleanup-mess-up ,cleanup-var))))
+         ((not ,cleanup-var) ,return-value)
+       ,@body)))
 
 ;;; Bind the IR1 context variables to the values associated with NODE,
 ;;; so that IR1 conversion can be done after the main conversion pass
@@ -988,64 +994,3 @@ specify bindings for printer control variables.")
         (nreverse (mapcar #'car *compiler-print-variable-alist*))
         (nreverse (mapcar #'cdr *compiler-print-variable-alist*))
       ,@forms)))
-
-#|
-In trying to figure out whether we can rectify the totally unobvious behavior
-that FUN-INFO-TRANSFORMS are tried in the reverse order of their definitions,
-it is helpful to understand when multiple transforms exist where more than
-one could be selected.
-In the absence of type overlap, it wouldn't matter what order they were attempted.
-Unfortunately there are plenty of cases where the _least_ _specific_ fun-type
-should be attempted first, which at least superfically makes no sense at all.
-
-(in-package sb-c)
-;; As a special case, delete any functions where there are 2 transforms
-;; and they can't possibly both apply. Just check the first arg.
-(defun domains-possibly-overlap (xforms)
-  (when (/= (length xforms) 2) ; lazy, just say yes
-    (return-from domains-possibly-overlap t))
-  (let ((t1 (transform-type (car xforms)))
-        (t2 (transform-type (cadr xforms))))
-    (unless (and (fun-type-p t1) (fun-type-p t2))
-      (return-from domains-possibly-overlap t))
-    (let ((req1 (car (fun-type-required t1)))
-          (req2 (car (fun-type-required t2))))
-      (unless (and req1 req2)
-        (return-from domains-possibly-overlap t))
-      (when (or (and (type= req1 (specifier-type 'single-float))
-                     (type= req2 (specifier-type 'double-float)))
-                (and (type= req1 (specifier-type 'double-float))
-                     (type= req2 (specifier-type 'single-float))))
-        (return-from domains-possibly-overlap nil))))
-  ;; The conservative answer is always T.
-  t)
-(let (list)
-  (do-all-symbols (s)
-    (dolist (name (list s `(setf ,s) `(cas ,s)))
-      (let* ((info (sb-int:info :function :info name))
-             (xforms (and info (sb-c::fun-info-transforms info))))
-        (when (and info
-                   (> (length xforms) 1)
-                   (not (assoc name list)))
-          (if (domains-possibly-overlap xforms)
-              (push (cons name xforms) list)
-              (format t "~&Nonoverlapping xforms on ~s~%" name))))))
-  (let ((*print-pretty* nil))
-    (dolist (x (sort (copy-list list) #'> :key (lambda (x) (length (cdr x))))
-               (format t "~s functions~%" (length list)))
-      (format t "~a~%" (car x))
-      (let ((i 0))
-        (dolist (xform (cdr x))
-          (incf i)
-          (format t " ~2d. ~s~%" i (sb-kernel:type-specifier
-                                  (sb-c::transform-type xform)))
-          (let* ((fun (sb-c::transform-function xform))
-                 (code (sb-kernel:fun-code-header fun))
-                 (info (sb-kernel:%code-debug-info code))
-                 (source (sb-c::compiled-debug-info-source info))
-                 (tlf (sb-c::compiled-debug-info-tlf-number info))
-                 (offs (sb-c::compiled-debug-info-char-offset info))
-                 (ns (sb-c::debug-source-namestring source))
-                 (doc (documentation (sb-c::transform-function xform) 'function)))
-            (format t "     ; ~a ~a ~d ~d~%" doc ns tlf offs)))))))
-|#
