@@ -70,7 +70,9 @@
 ;;; done on the basis of the primitive types of the operands, and the
 ;;; primitive type of a value is used to constrain the possible
 ;;; representations of that value.
-(defstruct (primitive-type (:copier nil))
+(defstruct (primitive-type (:constructor make-primitive-type (name scs specifier))
+                           (:copier nil)
+                           (:predicate nil))
   ;; the name of this PRIMITIVE-TYPE
   (name nil :type symbol :read-only t)
   ;; a list of the SC numbers for all the SCs that a TN of this type
@@ -132,9 +134,16 @@
 ;;; An IR2-BLOCK holds information about a block that is used during
 ;;; and after IR2 conversion. It is stored in the BLOCK-INFO slot for
 ;;; the associated block.
-(defstruct (ir2-block (:include block-annotation)
-                      (:constructor make-ir2-block (block))
+(defstruct (ir2-block (:constructor make-ir2-block (block))
                       (:copier nil))
+  ;; The IR1 block that this block is in the INFO for.
+  (block (missing-arg) :type cblock :read-only t)
+  ;; the next and previous block in emission order (not DFO). This
+  ;; determines which block we drop though to, and is also used to
+  ;; chain together overflow blocks that result from splitting of IR2
+  ;; blocks in lifetime analysis.
+  (next nil :type (or ir2-block null))
+  (prev nil :type (or ir2-block null))
   ;; the IR2-BLOCK's number, which differs from BLOCK's BLOCK-NUMBER
   ;; if any blocks are split. This is assigned by lifetime analysis.
   (number nil :type (or index null))
@@ -155,9 +164,6 @@
   ;; first.
   (start-stack () :type list)
   (end-stack () :type list)
-  ;; list of all lvars ever pushed onto the stack when control reaches
-  ;; the start of this block.
-  (stack-mess-up () :type list)
   ;; the first and last VOP in this block. If there are none, both
   ;; slots are null.
   (start-vop nil :type (or vop null))
@@ -220,6 +226,7 @@
 ;;; function result LVAR or that receive MVs.
 (defstruct (ir2-lvar
             (:constructor make-ir2-lvar (primitive-type))
+            (:predicate nil)
             (:copier nil))
   ;; If this is :DELAYED, then this is a single value LVAR for which
   ;; the evaluation of the use is to be postponed until the evaluation
@@ -246,7 +253,7 @@
   ;; restrictive than the tn-primitive-type of the value TN. This is
   ;; becase the value TN must hold any possible type that could be
   ;; computed (before type checking.) XXX
-  (primitive-type nil :type (or primitive-type null))
+  (primitive-type nil :type (or primitive-type null) :read-only t)
   ;; Locations used to hold the values of the LVAR. If the number of
   ;; values if fixed, then there is one TN per value. If the number of
   ;; values is unknown, then this is a two-list of TNs holding the
@@ -263,7 +270,8 @@
 
 ;;; An IR2-COMPONENT serves mostly to accumulate non-code information
 ;;; about the component being compiled.
-(defstruct (ir2-component (:copier nil))
+(defstruct (ir2-component (:copier nil)
+                          (:constructor make-ir2-component))
   ;; the counter used to allocate global TN numbers
   (global-tn-counter 0 :type index)
   ;; NORMAL-TNS is the head of the list of all the normal TNs that
@@ -306,8 +314,7 @@
   ;; constant pool. A non-immediate :CONSTANT TN with offset 0 refers
   ;; to the constant in element 0, etc. Normal constants are
   ;; represented by the placing the CONSTANT leaf in this vector. A
-  ;; load-time constant is distinguished by being a cons
-  ;; (KIND WHAT TN).
+  ;; load-time constant is distinguished by being a cons (KIND WHAT).
   ;; KIND is a keyword indicating how the constant is computed, and
   ;; WHAT is some context.
   ;;
@@ -329,11 +336,14 @@
   ;;    Is replaced with the result of executing the forms
   ;;    to compute <handle>.
   ;;
+  ;; as well as a few other forms of magic - see DUMP-CODE-OBJECT.
+  ;; (:coverage-marks )
+
   ;; A null entry in this vector is a placeholder for implementation
   ;; overhead that is eventually stuffed in somehow.
   ;; Prior to performing SORT-BOXED-CONSTANTS, the index 0 is reserved
   ;; to signify something (other than NIL) if stored as a TN-OFFSET,
-  ;; though nothing makes use of this at present.
+  ;; though nothing makes use of this at present. (Uh, what did I mean by this?)
   (constants (make-array 10 :fill-pointer 1 :adjustable t
                          :initial-element :ignore)
              :type vector :read-only t)
@@ -368,7 +378,9 @@
 ;;; since IR2 conversion may need to compile a forward reference. In
 ;;; this case the slots aren't actually initialized until entry
 ;;; analysis runs.
-(defstruct (entry-info (:copier nil))
+(defstruct (entry-info (:copier nil)
+                       (:predicate nil)
+                       (:constructor make-entry-info ()))
   ;; TN, containing closure (if needed) for this function in the home
   ;; environment.
   (closure-tn nil :type (or null tn))
@@ -394,7 +406,9 @@
 
 ;;; An IR2-ENVIRONMENT is used to annotate non-LET LAMBDAs with their
 ;;; passing locations. It is stored in the ENVIRONMENT-INFO.
-(defstruct (ir2-environment (:copier nil))
+(defstruct (ir2-environment (:copier nil)
+                            (:predicate nil)
+                            (:constructor make-ir2-environment (closure return-pc-pass)))
   ;; TN info for closed-over things within the function: an alist
   ;; mapping from NLX-INFOs and LAMBDA-VARs to TNs holding the
   ;; corresponding thing within this function
@@ -454,44 +468,49 @@
 ;;; A RETURN-INFO is used by GTN to represent the return strategy and
 ;;; locations for all the functions in a given TAIL-SET. It is stored
 ;;; in the TAIL-SET-INFO.
-(defstruct (return-info (:copier nil))
+(defstruct (return-info (:copier nil)
+                        (:predicate nil)
+                        (:constructor make-return-info (kind count primitive-types types
+                                                        &optional locations)))
   ;; The return convention used:
   ;; -- If :UNKNOWN, we use the standard return convention.
   ;; -- If :FIXED, we use the known-values convention.
-  (kind (missing-arg) :type (member :fixed :unknown :unboxed))
+  (kind (missing-arg) :type (member :fixed :unknown :unboxed) :read-only t)
   ;; the number of values returned, or :UNKNOWN if we don't know.
   ;; COUNT may be known when KIND is :UNKNOWN, since we may choose the
   ;; standard return convention for other reasons.
-  (count (missing-arg) :type (or index (member :unknown)))
+  (count (missing-arg) :type (or index (member :unknown)) :read-only t)
   ;; If count isn't :UNKNOWN, then this is a list of the
   ;; primitive-types of each value.
-  (primitive-types () :type list)
-  (types nil :type list)
+  (primitive-types () :type list :read-only t)
+  (types nil :type list :read-only t)
   ;; If kind is :FIXED, then this is the list of the TNs that we
   ;; return the values in.
-  (locations () :type list))
+  (locations () :type list :read-only t))
 (defprinter (return-info)
   kind
   count
   types
   locations)
 
-(defstruct (ir2-nlx-info (:copier nil))
+(defstruct (ir2-nlx-info (:copier nil)
+                         (:predicate nil)
+                         (:constructor make-ir2-nlx-info (home save-sp block-tn)))
   ;; If the kind is :ENTRY (a lexical exit), then in the home
   ;; environment, this holds a VALUE-CELL object containing the unwind
   ;; block pointer. In the other cases nobody directly references the
   ;; unwind-block, so we leave this slot null.
-  (home nil :type (or tn null))
+  (home nil :type (or tn null) :read-only t)
   ;; the saved control stack pointer
-  (save-sp nil :type (or tn null))
+  (save-sp nil :type (or tn null) :read-only t)
   ;; the list of dynamic state save TNs
   #-unbind-in-unwind
   (dynamic-state (list* (make-stack-pointer-tn)
                         (make-dynamic-state-tns))
                  :type list)
   ;; the target label for NLX entry
-  (target (gen-label) :type label)
-  (block-tn nil :type (or tn null)))
+  (target (gen-label) :type label :read-only t)
+  (block-tn nil :type (or tn null) :read-only t))
 (defprinter (ir2-nlx-info)
   home
   save-sp
@@ -505,7 +524,7 @@
 (defstruct (vop (:constructor make-vop (block node info args results))
                 (:copier nil))
   ;; VOP-INFO structure containing static info about the operation
-  (info nil :type vop-info)
+  (info nil :type vop-info :read-only t)
   ;; the IR2-BLOCK this VOP is in
   (block (missing-arg) :type ir2-block)
   ;; VOPs evaluated after and before this one. Null at the
@@ -515,8 +534,8 @@
   (prev nil :type (or vop null))
   ;; heads of the TN-REF lists for operand TNs, linked using the
   ;; ACROSS slot
-  (args nil :type (or tn-ref null))
-  (results nil :type (or tn-ref null))
+  (args nil :type (or tn-ref null) :read-only t)
+  (results nil :type (or tn-ref null) :read-only t)
   ;; head of the list of write refs for each explicitly allocated
   ;; temporary, linked together using the ACROSS slot
   (temps nil :type (or tn-ref null))
@@ -528,7 +547,7 @@
   ;; codegen. The meaning of this slot is totally dependent on the VOP.
   codegen-info
   ;; the node that generated this VOP, for keeping track of debug info
-  (node nil :type (or node null))
+  (node nil :type (or node null) :read-only t)
   ;; LOCAL-TN-BIT-VECTOR representing the set of TNs live after args
   ;; are read and before results are written. This is only filled in
   ;; when VOP-INFO-SAVE-P is non-null.
@@ -538,11 +557,12 @@
 ;;; to a TN. The information in TN-REFs largely determines how TNs are
 ;;; packed.
 (defstruct (tn-ref (:constructor make-tn-ref (tn write-p))
-                   (:copier nil))
+                   (:copier nil)
+                   (:predicate nil))
   ;; the TN referenced
   (tn (missing-arg) :type tn)
   ;; Is this is a write reference? (as opposed to a read reference)
-  (write-p nil :type boolean)
+  (write-p nil :type boolean :read-only t)
   ;; the link for a list running through all TN-REFs for this TN of
   ;; the same kind (read or write)
   (next nil :type (or tn-ref null))
@@ -725,7 +745,8 @@
   (optimizer nil :type (or null function (cons function symbol)))
   (optional-results nil :type list)
   move-vop-p
-  (after-sc-selection nil :type (or null function) :read-only t))
+  (after-sc-selection nil :type (or null function) :read-only t)
+  gc-barrier)
 (!set-load-form-method vop-info (:xc :target) :ignore-it)
 
 (declaim (inline vop-name))
@@ -794,7 +815,8 @@
 
 ;;; The SB structure represents the global information associated with
 ;;; a storage base.
-(defstruct (storage-base (:copier nil) (:conc-name sb-))
+(defstruct (storage-base (:copier nil) (:conc-name sb-)
+                         (:predicate nil))
   ;; name, for printing and reference
   (name nil :type symbol :read-only t)
   ;; the kind of storage base (which determines the packing
@@ -819,7 +841,8 @@
   ;; current-size must always be a multiple of this. It is assumed
   ;; to be a power of two.
   (size-alignment 1 :type index :read-only t))
-(defstruct (finite-sb (:copier nil) (:predicate nil) (:conc-name fsb-))
+(defstruct (finite-sb (:copier nil) (:predicate nil) (:conc-name fsb-)
+                      (:constructor make-finite-sb (conflicts always-live live-tns)))
   ;; the number of locations currently allocated in this SB
   (current-size 0 :type index)
   ;; the last location packed in, used by pack to scatter TNs to
@@ -956,7 +979,7 @@
 ;;;; TNs
 
 (defstruct (tn (:include sset-element)
-               (:constructor make-random-tn)
+               (:constructor make-random-tn (sc offset &aux (kind :normal)))
                (:constructor make-tn (number kind primitive-type sc))
                (:copier nil))
   ;; The kind of TN this is:
@@ -1090,6 +1113,8 @@
   ;; environment that the TN is live throughout.
   (environment nil :type (or environment null))
   ;; Used by pack-iterative
+  ;; Aliased TNs have :ALIAS there temporarily, to detected any extra
+  ;; writers.
   (vertex nil))
 
 (declaim (freeze-type tn))
@@ -1112,7 +1137,7 @@
             (:constructor make-global-conflicts (kind tn block number))
             (:copier nil))
   ;; the IR2-BLOCK that this structure represents the conflicts for
-  (block (missing-arg) :type ir2-block)
+  (block (missing-arg) :type ir2-block :read-only t)
   ;; thread running through all the GLOBAL-CONFLICTSs for BLOCK.
   (next-blockwise nil :type (or global-conflicts null))
   ;; the way that TN is used by BLOCK
@@ -1156,6 +1181,7 @@
 
 (defstruct (conditional-flags
             (:constructor make-conditional-flags (flags))
+            (:predicate nil)
             (:copier nil))
   flags)
 
