@@ -12,6 +12,10 @@
   :prefilter #'invert-4
   :printer #'print-ymmreg)
 
+(define-arg-type vvvv-reg
+  :prefilter #'invert-4
+  :printer #'print-reg)
+
 (define-arg-type ymm-reg-is4
   :printer #'print-ymmreg)
 
@@ -221,7 +225,9 @@
                   1)
                  ((xmm-register-p r)
                   0))))
-    (let ((l (cond (l)
+    (let ((l (cond ((eq l :from-thing)
+                    (xmm-size thing))
+                   (l)
                    ((xmm-register-p reg)
                     (xmm-size reg))
                    ((xmm-register-p thing)
@@ -347,17 +353,30 @@
            (emit-avx2-inst-imm segment dst src imm
                                #x66 ,opcode ,/i)))))
   (def vpslldq #x73 7)
-  (def vpsllw-imm #x71 6)               ; FIXME: get rid of that -IMM
-  (def vpslld-imm #x72 6)
-  (def vpsllq-imm #x73 6)
+  (def vpsrldq #x73 3))
 
-  (def vpsraw-imm #x71 4)
-  (def vpsrad-imm #x72 4)
+(macrolet
+    ((def (name opcode vopcode /i)
+       `(define-instruction ,name (segment dst src src2/imm)
+          ,@(avx2-inst-printer-list 'ymm-ymm-imm #x66 opcode
+                                    :more-fields `((/i ,/i)))
+          ,@(avx2-inst-printer-list 'ymm-ymm/mem #x66 vopcode :nds t)
+          (:emitter
+           (if (integerp src2/imm)
+               (emit-avx2-inst-imm segment dst src src2/imm
+                                   #x66 ,opcode ,/i)
+               (emit-avx2-inst segment src2/imm dst #x66 ,vopcode
+                               :vvvv src))))))
+  (def vpsllw #x71 #xf1 6)
+  (def vpslld #x72 #xf2 6)
+  (def vpsllq #x73 #xf3 6)
 
-  (def vpsrldq #x73 3)
-  (def vpsrlw-imm #x71 2)
-  (def vpsrld-imm #x72 2)
-  (def vpsrlq-imm #x73 2))
+  (def vpsraw #x71 #xe1 4)
+  (def vpsrad #x72 #xe2 4)
+
+  (def vpsrlw #x71 #xd1 2)
+  (def vpsrld #x72 #xd2 2)
+  (def vpsrlq #x73 #xd3 2))
 
 (macrolet ((def (name prefix opcode &optional (opcode-prefix #x0F))
              `(define-instruction ,name (segment dst src src2)
@@ -451,14 +470,7 @@
   (def vpmullw   #x66 #xd5)
   (def vpmuludq  #x66 #xf4)
   (def vpsadbw   #x66 #xf6)
-  (def vpsllw    #x66 #xf1)
-  (def vpslld    #x66 #xf2)
-  (def vpsllq    #x66 #xf3)
-  (def vpsraw    #x66 #xe1)
-  (def vpsrad    #x66 #xe2)
-  (def vpsrlw    #x66 #xd1)
-  (def vpsrld    #x66 #xd2)
-  (def vpsrlq    #x66 #xd3)
+
   (def vpsubb    #x66 #xf8)
   (def vpsubw    #x66 #xf9)
   (def vpsubd    #x66 #xfa)
@@ -517,13 +529,17 @@
   (def vaesdeclast #x66 #xdf #x0f38))
 
 ;;; Two arg instructions
-(macrolet ((def (name prefix opcode &optional (opcode-prefix #x0F))
+(macrolet ((def (name prefix opcode &optional (opcode-prefix #x0F) l)
              `(define-instruction ,name (segment dst src)
                 ,@(avx2-inst-printer-list 'ymm-ymm/mem prefix opcode
-                                          :opcode-prefix opcode-prefix)
+                                          :opcode-prefix opcode-prefix
+                                          :more-fields (and (eq l :from-thing)
+                                                            '((reg nil :type 'xmmreg))))
                 (:emitter
                  (emit-avx2-inst segment src dst ,prefix ,opcode
-                                 :opcode-prefix ,opcode-prefix)))))
+                                 :opcode-prefix ,opcode-prefix
+                                 ,@(and l
+                                       `(:l ,l)))))))
   ;; moves
   (def vmovshdup #xf3 #x16)
   (def vmovsldup #xf3 #x12)
@@ -540,13 +556,13 @@
   ;; conversion
   (def vcvtdq2pd #xf3 #xe6)
   (def vcvtdq2ps nil  #x5b)
-  (def vcvtpd2dq #xf2 #xe6)
-  (def vcvtpd2ps #x66 #x5a)
+  (def vcvtpd2dq #xf2 #xe6 #x0F :from-thing)
+  (def vcvtpd2ps #x66 #x5a #x0F :from-thing)
   (def vcvtps2dq #x66 #x5b)
   (def vcvtps2pd nil  #x5a)
   (def vcvtsd2ss #xf2 #x5a)
   (def vcvtss2sd #xf3 #x5a)
-  (def vcvttpd2dq #x66 #xe6)
+  (def vcvttpd2dq #x66 #xe6 #x0F :from-thing)
   (def vcvttps2dq #xf3 #x5b)
 
   (def vptest #x66 #x17 #x0f38)
@@ -576,7 +592,7 @@
                    'ymm-ymm/mem-imm prefix #x70
                    :printer '(:name :tab reg ", " reg/mem ", " imm))
                 (:emitter
-                 (emit-avx2-inst segment dst src ,prefix #x70
+                 (emit-avx2-inst segment src dst ,prefix #x70
                                  :remaining-bytes 1)
                  (emit-byte segment pattern)))))
   (def vpshufd  #x66)
@@ -995,18 +1011,21 @@
                                  :remaining-bytes 1)
                  (emit-byte segment imm)))))
   (def vpermpd #x66 #x01)
-  (def vpermpq #x66 #x00))
+  (def vpermq #x66 #x00))
 
-(define-instruction vpermps (segment dst src src2)
-  (:emitter
-   (emit-avx2-inst segment src2 dst #x66 #x16
-                   :opcode-prefix #x0f38
-                   :vvvv src
-                   :w 0 :l 1))
-  . #.(avx2-inst-printer-list 'ymm-ymm/mem #x66 #x16
-                              :w 0 :l 1
-                              :nds t
-                              :opcode-prefix #x0f38))
+(macrolet ((def (name op)
+             `(define-instruction ,name (segment dst src src2)
+                ,@(avx2-inst-printer-list 'ymm-ymm/mem #x66 op
+                                          :w 0 :l 1
+                                          :nds t
+                                          :opcode-prefix #x0f38)
+                (:emitter
+                 (emit-avx2-inst segment src2 dst #x66 ,op
+                                 :opcode-prefix #x0f38
+                                 :vvvv src
+                                 :w 0 :l 1)))))
+  (def vpermps #x16)
+  (def vpermd #x36))
 
 (macrolet ((def (name op op-imm)
              `(define-instruction ,name (segment dst src src2/imm)
@@ -1216,3 +1235,39 @@
                               :w 0
                               :opcode-prefix #x0f3a
                               :printer '(:name :tab reg/mem ", " reg ", " imm)))
+
+(define-instruction xsave (segment dst)
+  (:printer ext-reg/mem-no-width ((op '(#xae 4))))
+  (:emitter
+   (aver (not (register-p dst)))
+   (emit-prefixes segment dst nil :do-not-set)
+   (emit-byte segment #x0F)
+   (emit-byte segment #xAE)
+   (emit-ea segment dst 4)))
+
+(define-instruction xrstor (segment dst)
+  (:printer ext-reg/mem-no-width ((op '(#xae 5))))
+  (:emitter
+   (aver (not (register-p dst)))
+   (emit-prefixes segment dst nil :do-not-set)
+   (emit-byte segment #x0F)
+   (emit-byte segment #xAE)
+   (emit-ea segment dst 5)))
+
+(define-instruction-format (vex3-vex-gpr (+ 24 16) :include vex3
+                                                   :default-printer '(:name :tab reg ", " vvvv ", " reg/mem))
+  (op :field (byte 8 (+ 24 0)))
+  (vvvv :type 'vvvv-reg)
+  (reg/mem :fields (list (byte 2 (+ 24 14)) (byte 3 (+ 24 8))) :type 'reg/mem)
+  (reg :field (byte 3 (+ 24 11)) :type 'reg))
+
+
+(define-instruction mulx (segment &prefix prefix hi lo src)
+  (:emitter
+   (emit-avx2-inst segment src hi #xF2 #xF6 :opcode-prefix #x0f38 :vvvv lo
+                                            :w (ecase (pick-operand-size prefix lo src)
+                                                 (:qword 1)
+                                                 (:dword 0))))
+  . #.(avx2-inst-printer-list 'vex-gpr #xF2 #xF6
+                              :nds t
+                              :opcode-prefix #x0f38))
