@@ -24,14 +24,12 @@
   (descriptor-reg) (sap-reg))
 
 ;;; Move an untagged SAP to a tagged representation.
-(define-vop (move-from-sap)
+(define-allocator (move-from-sap)
   (:args (sap :scs (sap-reg) :to :result))
   (:results (res :scs (descriptor-reg) :from :argument))
-  #+gs-seg (:temporary (:sc unsigned-reg :offset 15) thread-tn)
   (:note "SAP to pointer coercion")
-  (:node-var node)
   (:generator 20
-    (alloc-other sap-widetag sap-size res node nil thread-tn)
+    (alloc-other sap-widetag sap-size res)
     (storew sap res sap-pointer-slot other-pointer-lowtag)))
 (define-move-vop move-from-sap :move
   (sap-reg) (descriptor-reg))
@@ -140,7 +138,26 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
 /// and we store the shadow _before_ the app store."
 |#
 
-(defun emit-sap-ref (size insn modifier result ea node vop temp)
+(defun offset-needs-temp (offset)
+  (if (or (sc-is offset signed-reg)
+          (typep (tn-value offset) '(signed-byte 32)))
+      nil
+      t))
+
+(defun sap+offset-to-ea (sap offset temp)
+  (multiple-value-bind (disp index)
+      (cond ((sc-is offset signed-reg)
+             (values 0 offset))
+            ((typep (tn-value offset) '(signed-byte 32))
+             (values (tn-value offset) nil))
+            (t
+             ;; Seriously who uses immediate offsets exceeding an int32?
+             (aver temp)
+             (inst mov temp (tn-value offset))
+             (values 0 temp)))
+    (ea disp sap index)))
+
+(defun emit-sap-ref (size insn modifier result sap offset node vop temp)
   (declare (ignorable node size vop temp))
   (cond
    ;; MSAN as implemented can't correctly interact with the C sanitizer because
@@ -150,7 +167,7 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
    ;; a 1, 2, 4, or 8-byte load. Unfortunately that load could spuriously report
    ;; that it was uninitialized if only _part_ was uninitialized, but we didn't
    ;; want to examine the uninitialized bits. The problem is, it doesn't know that.
-   ;; Even though DEFINE-ALIEN-TYPE conveys the sematics of sub-byte fields, the
+   ;; Even though DEFINE-ALIEN-TYPE conveys the semantics of sub-byte fields, the
    ;; compiler is itself unable to extract partial bytes.
    ;; (You get "cannot extract 4-bit integers" for example)
    ;; So unfortunately, because the IR can't convey what we mean, we have to do
@@ -161,7 +178,7 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
     (aver (not (location= temp result)))
     (inst lea temp ea)
     (sb-assem:inst* insn modifier result (ea temp))
-    (inst xor temp (thread-slot-ea thread-msan-xor-constant-slot))
+    (inst xor temp (static-constant-ea msan-xor-constant))
     ;; Per the documentation, shadow is tested _after_
     (let ((mask (sb-c::masked-memory-load-p vop))
           (good (gen-label))
@@ -190,43 +207,41 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
         (inst byte (logior (ash (tn-offset result) 2) scale)))
       (emit-label good)))
    (t
-    (sb-assem:inst* insn modifier result ea))))
+    (let ((ea (sap+offset-to-ea sap offset temp)))
+      (sb-assem:inst* insn modifier result ea)))))
 
-(defun emit-sap-set (size ea value temp)
+(defun emit-sap-set (size sap offset value temp
+                          &aux (ea (sap+offset-to-ea sap offset temp)))
   #+linux
   (when (sb-c:msan-unpoison sb-c:*compilation*)
     (inst lea temp ea)
-    (inst xor temp (thread-slot-ea thread-msan-xor-constant-slot))
+    (inst xor temp (static-constant-ea msan-xor-constant))
     (inst mov size (ea temp) 0))
   (when (sc-is value constant immediate)
     (cond ((plausible-signed-imm32-operand-p (tn-value value))
            (setq value (tn-value value)))
           (t
+           (aver (not (offset-needs-temp offset)))
            (inst mov temp (tn-value value))
            (setq value temp))))
   (inst mov size ea value))
 
-(defun emit-cas-sap-ref (size sap offset oldval newval result rax temp)
-  (multiple-value-bind (disp index)
-      (cond ((sc-is offset signed-reg)
-             (values 0 offset))
-            ((typep (tn-value offset) '(signed-byte 32))
-             (values (tn-value offset) nil))
-            (t
-             (inst mov temp (tn-value offset))
-             (values 0 temp)))
+(defun emit-cas-sap-ref (size signedp sap offset oldval newval result rax temp)
+  (let ((ea (sap+offset-to-ea sap offset temp)))
     (cond ((sc-is oldval immediate constant)
            (inst mov rax (tn-value oldval)))
           ((not (location= oldval rax))
            (inst mov (if (eq size :qword) :qword :dword) rax oldval)))
-    (inst cmpxchg size :lock (ea disp sap index) newval)
-    (unless (location= result rax)
-      (inst mov (if (eq size :qword) :qword :dword) result rax))))
+    (inst cmpxchg size :lock ea newval)
+    (cond ((and signedp (neq size :qword)) ; MOVSX is required no matter what
+           (inst movsx `(,size :qword) result rax))
+          ((not (location= result rax))
+           (inst mov (if (eq size :qword) :qword :dword) result rax)))))
 
-;;; TODO: these should be refactored so that there is only one vop for any given
+;;; Note: these could potentially be arranged to have only one vop per
 ;;; result storage class. In particular, sap-ref-{8,16,32} can all produce tagged-num.
 ;;; The vop can examine the node to see which function it translates
-;;; and select the appropriate modifier to movzx or movsx.
+;;; and select the appropriate modifier to movzx or movsx. I'm unsure if that's better.
 (macrolet ((def-system-ref-and-set (ref-name
                                     set-name
                                     ref-insn
@@ -244,9 +259,7 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
                                  size
                                  `(,size ,(if (eq ref-insn 'movzx) :dword :qword)))))
                `(progn
-                  ,@(when (member ref-name '(sap-ref-8 sap-ref-16 sap-ref-32 sap-ref-64
-                                             signed-sap-ref-64
-                                             sap-ref-lispobj sap-ref-sap))
+                  ,@(when (implements-cas-sap-ref ref-name)
                       `((define-vop (,(symbolicate "CAS-" ref-name))
                           (:translate (cas ,ref-name))
                           (:policy :fast-safe)
@@ -261,37 +274,24 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
                                        :from (:argument 0) :to :result) rax)
                           (:temporary (:sc unsigned-reg) temp)
                           (:generator 3
-                            (emit-cas-sap-ref ',size sap offset oldval newval result rax temp)))))
+                            (emit-cas-sap-ref ',size ,(eq sc 'signed-reg)
+                                              sap offset oldval newval result rax temp)))))
                   (define-vop (,ref-name)
                     (:translate ,ref-name)
                     (:policy :fast-safe)
                     (:args (sap :scs (sap-reg))
-                           (offset :scs (signed-reg)))
+                           (offset :scs (signed-reg immediate)))
                     (:arg-types system-area-pointer signed-num)
                     (:results (result :scs (,sc)))
                     (:result-types ,type)
                     (:node-var node)
                     (:vop-var vop)
-                    ;; this temp has to be wired because the uninitialized-load-trap handler
-                    ;; looks in RAX to get the poisoned address.
-                    ;; We should have a different variant of this reffer for msan or no msan
-                    ;; to avoid wasting a register that is not needed.
-                    (:temporary (:sc unsigned-reg :offset rax-offset) temp)
+                    (:temporary (:sc unsigned-reg :offset rax-offset
+                                 :unused-if (and (not (sb-c:msan-unpoison sb-c:*compilation*))
+                                                 (not (offset-needs-temp offset))))
+                                temp)
                     (:generator 3 (emit-sap-ref ,size ',ref-insn
-                                                ',modifier result (ea sap offset) node vop temp)))
-                  (define-vop (,(symbolicate ref-name "-C"))
-                    (:translate ,ref-name)
-                    (:policy :fast-safe)
-                    (:args (sap :scs (sap-reg)))
-                    (:arg-types system-area-pointer (:constant (signed-byte 32)))
-                    (:info offset)
-                    (:results (result :scs (,sc)))
-                    (:result-types ,type)
-                    (:node-var node)
-                    (:vop-var vop)
-                    (:temporary (:sc unsigned-reg :offset rax-offset) temp)
-                    (:generator 2 (emit-sap-ref ,size ',ref-insn
-                                                ',modifier result (ea offset sap) node vop temp)))
+                                                ',modifier result sap offset node vop temp)))
                   (define-vop (,set-name)
                     (:translate ,set-name)
                     (:policy :fast-safe)
@@ -299,19 +299,11 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
                            (sap :scs (sap-reg))
                            (offset :scs (signed-reg)))
                     (:arg-types ,type system-area-pointer signed-num)
+                    ;; In theory TEMP could assist with loading either OFFSET or VALUE, but
+                    ;; there is an AVER that it doesn't actually get used for both.
                     (:temporary (:sc unsigned-reg) temp)
                     (:generator 5
-                      (emit-sap-set ,size (ea sap offset) value temp)))
-                  (define-vop (,(symbolicate set-name "-C"))
-                    (:translate ,set-name)
-                    (:policy :fast-safe)
-                    (:args (value :scs ,value-scs)
-                           (sap :scs (sap-reg)))
-                    (:arg-types ,type system-area-pointer (:constant (signed-byte 32)))
-                    (:info offset)
-                    (:temporary (:sc unsigned-reg) temp)
-                    (:generator 4
-                      (emit-sap-set ,size (ea offset sap) value temp)))))))
+                      (emit-sap-set ,size sap offset value temp)))))))
 
   (def-system-ref-and-set sap-ref-8 %set-sap-ref-8 movzx
     unsigned-reg positive-fixnum :byte)
@@ -336,45 +328,51 @@ https://llvm.org/doxygen/MemorySanitizer_8cpp.html
 
 ;;;; SAP-REF-SINGLE and SAP-REF-DOUBLE
 
-(macrolet ((def-system-ref-and-set (ref-fun res-sc res-type insn
+(macrolet ((def-system-ref-and-set (ref-fun res-sc res-type insn cas
                                             &aux (set-fun (symbolicate "%SET-" ref-fun)))
              `(progn
                 (define-vop (,ref-fun)
                   (:translate ,ref-fun)
                   (:policy :fast-safe)
                   (:args (sap :scs (sap-reg))
-                         (offset :scs (signed-reg)))
+                         (offset :scs (signed-reg immediate)))
                   (:arg-types system-area-pointer signed-num)
                   (:results (result :scs (,res-sc)))
                   (:result-types ,res-type)
-                  (:generator 5 (inst ,insn result (ea sap offset))))
-                (define-vop (,(symbolicate ref-fun "-C"))
-                  (:translate ,ref-fun)
-                  (:policy :fast-safe)
-                  (:args (sap :scs (sap-reg)))
-                  (:arg-types system-area-pointer (:constant (signed-byte 32)))
-                  (:info offset)
-                  (:results (result :scs (,res-sc)))
-                  (:result-types ,res-type)
-                  (:generator 4 (inst ,insn result (ea offset sap))))
+                  (:generator 1 (inst ,insn result (sap+offset-to-ea sap offset nil))))
                 (define-vop (,set-fun)
                   (:translate ,set-fun)
                   (:policy :fast-safe)
                   (:args (value :scs (,res-sc immediate))
                          (sap :scs (sap-reg))
-                         (offset :scs (signed-reg)))
+                         (offset :scs (signed-reg immediate)))
                   (:arg-types ,res-type system-area-pointer signed-num)
-                  (:generator 5 (inst ,insn (ea sap offset) value)))
-                (define-vop (,(symbolicate set-fun "-C"))
-                  (:translate ,set-fun)
+                  (:generator 1 (inst ,insn (sap+offset-to-ea sap offset nil) value)))
+                (define-vop (,(symbolicate "CAS-" ref-fun))
+                  (:translate (cas ,ref-fun))
                   (:policy :fast-safe)
-                  (:args (value :scs (,res-sc))
-                         (sap :scs (sap-reg)))
-                  (:arg-types ,res-type system-area-pointer (:constant (signed-byte 32)))
-                  (:info offset)
-                  (:generator 4 (inst ,insn (ea offset sap) value))))))
-  (def-system-ref-and-set sap-ref-single single-reg single-float movss)
-  (def-system-ref-and-set sap-ref-double double-reg double-float movsd))
+                  ;; old and new could directly accept descriptor-reg
+                  ;; but I doubt that CAS on floats sees enough usage to care.
+                  (:args (oldval :scs (,res-sc))
+                         (newval :scs (,res-sc))
+                         (sap :scs (sap-reg))
+                         (offset :scs (signed-reg immediate)))
+                  (:arg-types ,res-type ,res-type system-area-pointer signed-num)
+                  (:results (result :scs (,res-sc)))
+                  (:result-types ,res-type)
+                  (:temporary (:sc unsigned-reg :offset rax-offset) rax)
+                  (:temporary (:sc unsigned-reg) newval-temp)
+                  (:generator 3 ,@cas)))))
+  (def-system-ref-and-set sap-ref-single single-reg single-float movss
+    ((inst movd rax oldval)
+     (inst movd newval-temp newval)
+     (inst cmpxchg :dword :lock (sap+offset-to-ea sap offset nil) newval-temp)
+     (inst movd result rax)))
+  (def-system-ref-and-set sap-ref-double double-reg double-float movsd
+    ((inst movq rax oldval)
+     (inst movq newval-temp newval)
+     (inst cmpxchg :lock (sap+offset-to-ea sap offset nil) newval-temp)
+     (inst movq result rax))))
 
 ;;; noise to convert normal lisp data objects into SAPs
 
