@@ -706,13 +706,10 @@
 ;;; In tail call with fixed arguments, the passing locations are
 ;;; passed as a more arg, but there is no new-FP, since the arguments
 ;;; have been set up in the current frame.
-(macrolet ((define-full-call (name named return variable)
-            (aver (not (and variable (eq return :tail))))
-            `(define-vop (,name
-                          ,@(when (eq return :unknown)
-                              '(unknown-values-receiver)))
-               (:args
-               ,@(unless (eq return :tail)
+(defmacro define-full-call (name named return variable)
+  (aver (not (and variable (eq return :tail))))
+  `(define-vop (,name ,@(when (eq return :unknown) '(unknown-values-receiver)))
+     (:args    ,@(unless (eq return :tail)
                    '((new-fp :scs (any-reg) :to (:argument 1))))
                ,@(unless (eq named :direct)
                    '((fun :scs (descriptor-reg control-stack)
@@ -727,21 +724,19 @@
                ,@(when (eq return :fixed)
                '((:results (values :more t))))
 
-               (:save-p ,(if (eq return :tail) :compute-only t))
+     (:save-p ,(if (eq return :tail) :compute-only t))
 
-               ,@(unless (or (eq return :tail) variable)
-               '((:move-args :full-call)))
+     ,@(unless (or (eq return :tail) variable)
+         '((:move-args :full-call)))
 
-               (:vop-var vop)
-               (:info
-               ,@(unless (or variable (eq return :tail)) '(arg-locs))
+     (:vop-var vop)
+     (:info    ,@(unless (or variable (eq return :tail)) '(arg-locs))
                ,@(unless variable '(nargs))
                ,@(when (eq named :direct) '(fun))
                ,@(when (eq return :fixed) '(nvals))
                step-instrumenting)
 
-               (:ignore
-               ,@(unless (or variable (eq return :tail)) '(arg-locs))
+     (:ignore  ,@(unless (or variable (eq return :tail)) '(arg-locs))
                ,@(unless variable '(args))
                ,@(when (eq named :direct) '(step-instrumenting)))
 
@@ -751,69 +746,58 @@
                ;; with the real function and invoke the real function
                ;; for closures. Non-closures do not need this value,
                ;; so don't care what shows up in it.
-               ,@(unless (eq named :direct)
-                   '((:temporary
-                      (:sc descriptor-reg
-                       :offset eax-offset
-                       :from (:argument 0)
-                       :to :eval)
-                      eax)))
+     ,@(unless (eq named :direct)
+         '((:temporary
+            (:sc descriptor-reg :offset eax-offset :from (:argument 0) :to :eval) eax)))
 
-               ;; We pass the number of arguments in ECX.
-               (:temporary (:sc unsigned-reg :offset ecx-offset :to :eval) ecx)
+     ;; We pass the number of arguments in ECX.
+     (:temporary (:sc unsigned-reg :offset ecx-offset :to :eval) ecx)
 
                ;; With variable call, we have to load the
                ;; register-args out of the (new) stack frame before
                ;; doing the call. Therefore, we have to tell the
                ;; lifetime stuff that we need to use them.
-               ,@(when variable
-                   (mapcar (lambda (name offset)
-                             `(:temporary (:sc descriptor-reg
-                                               :offset ,offset
-                                               :from (:argument 0)
-                                               :to :eval)
-                                          ,name))
-                           *register-arg-names* *register-arg-offsets*))
+     ,@(when variable
+         (mapcar (lambda (name offset)
+                   `(:temporary (:sc descriptor-reg
+                                 :offset ,offset
+                                 :from (:argument 0)
+                                 :to :eval)
+                                ,name))
+                 register-arg-names *register-arg-offsets*))
 
-               ,@(when (eq return :tail)
-                   '((:temporary (:sc unsigned-reg
-                                      :from (:argument 1)
-                                      :to (:argument 2))
-                                 old-fp-tmp)))
-               ,@(unless (eq return :tail)
-                   '((:node-var node)))
+     ,@(when (eq return :tail)
+         '((:temporary (:sc unsigned-reg :from (:argument 1) :to (:argument 2))
+            old-fp-tmp)))
+     ,@(unless (eq return :tail) '((:node-var node)))
 
-               (:generator ,(+ (if named 5 0)
-                               (if variable 19 1)
-                               (if (eq return :tail) 0 10)
-                               15
-                               (if (eq return :unknown) 25 0))
+     (:generator ,(+ (if named 5 0)
+                     (if variable 19 1)
+                     (if (eq return :tail) 0 10)
+                     15
+                     (if (eq return :unknown) 25 0))
 
                ;; This has to be done before the frame pointer is
                ;; changed! EAX stores the 'lexical environment' needed
                ;; for closures.
-               ,@(unless (eq named :direct)
-                   '((move eax fun)))
-               ,@(if variable
+       ,@(unless (eq named :direct) '((move eax fun)))
+
+       ,@(if variable
                      ;; For variable call, compute the number of
                      ;; arguments and move some of the arguments to
                      ;; registers.
-                     (collect ((noise))
-                              ;; Compute the number of arguments.
-                              (noise '(inst mov ecx new-fp))
-                              (noise '(inst sub ecx esp-tn))
+             `((inst mov ecx new-fp)
+               (inst sub ecx esp-tn)
                               ;; Move the necessary args to registers,
                               ;; this moves them all even if they are
                               ;; not all needed.
-                              (loop
-                               for name in *register-arg-names*
-                               for index downfrom -1
-                               do (noise `(loadw ,name new-fp ,index)))
-                              (noise))
-                   '((if (zerop nargs)
-                         (inst xor ecx ecx)
-                       (inst mov ecx (fixnumize nargs)))))
-               ,@(cond ((eq return :tail)
+               ,@(loop for name in register-arg-names
+                       for index downfrom -1
+                       collect `(loadw ,name new-fp ,index)))
+             '((if (zerop nargs)
+                   (inst xor ecx ecx)
+                   (inst mov ecx (fixnumize nargs)))))
+       ,@(cond ((eq return :tail)
                         '(;; Python has figured out what frame we should
                           ;; return to so might as well use that clue.
                           ;; This seems really important to the
@@ -849,7 +833,7 @@
                           ;; return-pc so that it looks like we CALLed
                           ;; despite the fact that we are going to JMP.
                           (inst push return-pc)))
-                       (t
+               (t
                         ;; For non-tail call, we have to save our
                         ;; frame pointer and install the new frame
                         ;; pointer. We can't load stack tns after this
@@ -876,47 +860,42 @@
                                   (frame-word-offset ocfp-save-offset))
 
                           (move ebp-tn new-fp))))  ; NB - now on new stack frame.
-               ,@(unless (eq named :direct)  ;; handle-single-step-around-trap can't handle it
-                   `((when step-instrumenting
-                       (emit-single-step-test)
-                       (inst jmp :eq DONE)
-                       (inst break single-step-around-trap))))
-               DONE
+       ,@(unless (eq named :direct)  ;; handle-single-step-around-trap can't handle it
+           `((when step-instrumenting
+               (emit-single-step-test)
+               (inst jmp :eq DONE)
+               (inst break single-step-around-trap))))
+       DONE
 
-               (note-this-location vop :call-site)
+       (note-this-location vop :call-site)
 
-               (inst ,(if (eq return :tail) 'jmp 'call)
-                     ,(case named
-                        (:direct
-                         '(make-ea :dword :disp
-                           (+ nil-value (static-fun-offset fun))))
-                        ((t)
-                         '(object-slot-ea eax fdefn-raw-addr-slot
-                           other-pointer-lowtag))
-                        ((nil)
-                         '(object-slot-ea eax closure-fun-slot
-                           fun-pointer-lowtag))))
-               ,@(ecase return
-                   (:fixed
-                    '((default-unknown-values vop values nvals node)))
-                   (:unknown
-                    '((note-this-location vop :unknown-return)
-                      (receive-unknown-values values-start nvals start count
-                                              node)))
-                   (:tail))))))
+       (inst ,(if (eq return :tail) 'jmp 'call)
+         ,(case named
+            (:direct
+             '(make-ea :dword :disp (+ nil-value (static-fun-offset fun))))
+            ((t)
+             '(object-slot-ea eax fdefn-raw-addr-slot other-pointer-lowtag))
+            ((nil)
+             '(object-slot-ea eax closure-fun-slot fun-pointer-lowtag))))
+       ,@(ecase return
+           (:fixed '((default-unknown-values vop values nvals node)))
+           (:unknown
+            '((note-this-location vop :unknown-return)
+              (receive-unknown-values values-start nvals start count node)))
+           (:tail)))))
 
-  (define-full-call call nil :fixed nil)
-  (define-full-call call-named t :fixed nil)
-  (define-full-call static-call-named :direct :fixed nil)
-  (define-full-call multiple-call nil :unknown nil)
-  (define-full-call multiple-call-named t :unknown nil)
-  (define-full-call static-multiple-call-named :direct :unknown nil)
-  (define-full-call tail-call nil :tail nil)
-  (define-full-call tail-call-named t :tail nil)
-  (define-full-call static-tail-call-named :direct :tail nil)
+(define-full-call call nil :fixed nil)
+(define-full-call call-named t :fixed nil)
+(define-full-call static-call-named :direct :fixed nil)
+(define-full-call multiple-call nil :unknown nil)
+(define-full-call multiple-call-named t :unknown nil)
+(define-full-call static-multiple-call-named :direct :unknown nil)
+(define-full-call tail-call nil :tail nil)
+(define-full-call tail-call-named t :tail nil)
+(define-full-call static-tail-call-named :direct :tail nil)
 
-  (define-full-call call-variable nil :fixed t)
-  (define-full-call multiple-call-variable nil :unknown t))
+(define-full-call call-variable nil :fixed t)
+(define-full-call multiple-call-variable nil :unknown t)
 
 ;;; This is defined separately, since it needs special code that BLT's
 ;;; the arguments down. All the real work is done in the assembly
