@@ -202,15 +202,16 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
                 (slot-definition-location slot)))))
     (setf *standard-slot-locations* new)))
 
-(defun maybe-update-standard-slot-locations (class)
-  (when (and (eq **boot-state** 'complete)
-             (memq (class-name class) +standard-classes+))
-    (compute-standard-slot-locations)))
-
 (defun standard-slot-value (object slot-name class)
   (declare (notinline standard-instance-access
                       funcallable-standard-instance-access))
-  (let ((location (gethash (cons class slot-name) *standard-slot-locations*)))
+  ;; I'm sure there's a super easy way to feed the mix of the CLASS and SLOT-NAME
+  ;; hashes into a perfect hash fun, but this function seems never to be called except
+  ;; by MOP some tests. Therefore I don't care to improve it beyond the avoidance
+  ;; of 1 cons operation.
+  (let* ((key (cons class slot-name))
+         (location (gethash key *standard-slot-locations*)))
+    (declare (dynamic-extent key))
     (if location
         (let ((value (if (funcallable-instance-p object)
                          (funcallable-standard-instance-access object location)
@@ -1118,10 +1119,17 @@ Except see also BREAK-VICIOUS-METACIRCLE.  -- CSR, 2003-05-28
               (index (standard-slot-value/eslotd slotd 'location))
               (type (gf-info-simple-accessor-type arg-info)))
           (when (and method
-                     (subtypep (ecase accessor-type
-                                 ((reader) (car classes))
-                                 ((writer) (cadr classes)))
-                               class))
+                     (let ((method-class (ecase accessor-type
+                                           ((reader) (car classes))
+                                           ((writer) (cadr classes)))))
+                       (or (eq method-class class)
+                           ;; SUBTYPEP doesn't work because it calls the CLASS-WRAPPER GF.
+                           (block nil
+                             (sb-kernel::do-subclassoids ((subclassoid layout)
+                                                          (layout-classoid (standard-slot-value/class class 'wrapper)))
+                               (declare (ignore layout))
+                               (when (eq method-class (classoid-pcl-class subclassoid))
+                                 (return t)))))))
             (return-from break-vicious-metacircle
               (values index (list method) type index)))))))
   (error "~@<vicious metacircle:  The computation of an ~
