@@ -32,7 +32,7 @@
 #include "lispregs.h"
 #include "runtime.h"
 #include "genesis/static-symbols.h"
-#include "genesis/fdefn.h"
+#include "genesis/symbol.h"
 
 #include <errno.h>
 
@@ -64,90 +64,6 @@ int personality (unsigned long);
 #include <unistd.h>
 #include <errno.h>
 
-#ifdef MUTEX_EVENTRECORDING
-#define MAXEVENTS 200
-static struct {
-    struct thread* th;
-    struct timespec ts;
-    char *label;
-    char *mutex_name;
-    sword_t timeout;
-} events[MAXEVENTS];
-static int record_mutex_events;
-static int eventcount;
-
-void lisp_mutex_event(char *string) {
-    if (record_mutex_events) {
-        int id = __sync_fetch_and_add(&eventcount, 1);
-        if (id >= MAXEVENTS) lose("event buffer overflow");
-        clock_gettime(CLOCK_REALTIME, &events[id].ts);
-        events[id].th = get_sb_vm_thread();
-        events[id].label = string;
-        events[id].mutex_name = 0;
-        events[id].timeout = -1;
-    }
-}
-void lisp_mutex_event1(char *string, char *string2) {
-    if (record_mutex_events) {
-        int id = __sync_fetch_and_add(&eventcount, 1);
-        if (id >= MAXEVENTS) lose("event buffer overflow");
-        clock_gettime(CLOCK_REALTIME, &events[id].ts);
-        events[id].th = get_sb_vm_thread();
-        events[id].label = string;
-        events[id].mutex_name = string2;
-        events[id].timeout = -1;
-    }
-}
-void lisp_mutex_event2(char *string, char *string2, uword_t usec) {
-    if (record_mutex_events) {
-        int id = __sync_fetch_and_add(&eventcount, 1);
-        if (id >= MAXEVENTS) lose("event buffer overflow");
-        clock_gettime(CLOCK_REALTIME, &events[id].ts);
-        events[id].th = get_sb_vm_thread();
-        events[id].label = string;
-        events[id].mutex_name = string2;
-        events[id].timeout = usec;
-    }
-}
-void lisp_mutex_start_eventrecording() {
-    eventcount = 0;
-    record_mutex_events = 1;
-}
-void lisp_mutex_done_eventrecording() {
-    record_mutex_events = 0;
-    int i;
-    fprintf(stderr, "event log:\n");
-    struct timespec basetime = events[0].ts;
-    for(i=0; i<eventcount;++i) {
-        struct thread *th = events[i].th;
-        struct thread_instance *ti = (void*)native_pointer(th->lisp_thread);
-        struct timespec rel_time = events[i].ts;
-        rel_time.tv_sec -= basetime.tv_sec;
-        rel_time.tv_nsec -= basetime.tv_nsec;
-        if (rel_time.tv_nsec<0) rel_time.tv_nsec += 1000 * 1000 * 1000, rel_time.tv_sec--;
-        lispobj threadname = ti->_name;
-        if (events[i].timeout >= 0) // must also have mutex_name in this case
-            fprintf(stderr, "[%d.%09ld] %s: %s '%s' timeout %ld\n",
-                    (int)rel_time.tv_sec, rel_time.tv_nsec,
-                    vector_sap(threadname),
-                    events[i].label, events[i].mutex_name, events[i].timeout);
-        else if (events[i].mutex_name)
-            fprintf(stderr, "[%d.%09ld] %s: %s '%s'\n",
-                    (int)rel_time.tv_sec, rel_time.tv_nsec,
-                    vector_sap(threadname), events[i].label, events[i].mutex_name);
-        else
-            fprintf(stderr, "[%d.%09ld] %s: %s\n",
-                    (int)rel_time.tv_sec, rel_time.tv_nsec,
-                    vector_sap(threadname), events[i].label);
-    }
-    fprintf(stderr, "-----\n");
-}
-#else
-#define lisp_mutex_event(x)
-#define lisp_mutex_event1(x,y)
-#define lisp_mutex_event2(x,y,z)
-#endif
-
 /* values taken from the kernel's linux/futex.h.  This header file
    doesn't exist in userspace, which is our excuse for not grovelling
    them automatically */
@@ -215,21 +131,14 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
   struct timespec timeout;
   int t;
 
-#ifdef MUTEX_EVENTRECORDING
-    struct lispmutex* m = (void*)((char*)lock_word - offsetof(struct lispmutex,uw_state));
-    char *name = m->name != NIL ? vector_sap(m->name) : "(unnamed)";
-#endif
   if (sec<0) {
-      lisp_mutex_event1("start futex wait", name);
       t = sys_futex(lock_word, futex_wait_op(), oldval, 0);
   }
   else {
       timeout.tv_sec = sec;
       timeout.tv_nsec = usec * 1000;
-      lisp_mutex_event2("start futex timedwait", name, usec);
       t = sys_futex(lock_word, futex_wait_op(), oldval, &timeout);
   }
-  lisp_mutex_event1("back from sys_futex", name);
   if (t==0)
       return 0;
   else if (errno==ETIMEDOUT)
@@ -244,11 +153,6 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
 int
 futex_wake(int *lock_word, int n)
 {
-#ifdef MUTEX_EVENTRECORDING
-    struct lispmutex* m = (void*)((char*)lock_word - offsetof(struct lispmutex,uw_state));
-    char *name = m->name != NIL ? vector_sap(m->name) : "(unnamed)";
-    lisp_mutex_event1("waking futex", name);
-#endif
     return sys_futex(lock_word, futex_wake_op(),n,0);
 }
 #endif
