@@ -9,10 +9,6 @@
  * files for more information.
  */
 
-#ifndef GENCGC_IS_PRECISE
-#error "GENCGC_IS_PRECISE must be #defined as 0 or 1"
-#endif
-
 /* Use AVX2 versions of code when we can, since blasting bytes faster
  * is always nice.
  * If used more widely, we should put these in runtime.h */
@@ -98,9 +94,9 @@ struct page {
      */
 #if CONDENSED_PAGE_TABLE
     // The low bit of the offset indicates the scale factor:
-    // 0 = double-lispwords, 1 = gc cards. Large objects are card-aligned,
+    // 0 = double-lispwords, 1 = gc cards. Large objects are page-aligned,
     // and this representation allows for a 32TB contiguous block using 32K
-    // card size. Larger allocations will have pages that can't directly
+    // page size. Larger allocations will have pages that can't directly
     // store the full offset. That has to be dealt with by the accessor.
     unsigned int scan_start_offset_;
 #else
@@ -329,7 +325,11 @@ extern void gc_close_collector_regions(int);
 /* The various sorts of pointer swizzling in SBCL. */
 enum source {
   SOURCE_NORMAL,
+#ifdef LISP_FEATURE_LINKAGE_SPACE
+  // avoid a warning from some C compilers that LINKAGE_CELL is not handled
+  // in "switch (source_type)"
   SOURCE_LINKAGE_CELL,
+#endif
   SOURCE_ZERO_TAG,              /* code, lflist */
   SOURCE_CLOSURE,
   SOURCE_FDEFN_RAW
@@ -363,6 +363,38 @@ gc_general_alloc(struct alloc_region* region, sword_t nbytes, int page_type)
 lispobj copy_potential_large_object(lispobj object, sword_t nwords,
                                    struct alloc_region*, int page_type);
 
+#define compacting_p() (from_space>=0)
+
+#define page_single_obj_p(page) ((page_table[page].type & SINGLE_OBJECT_FLAG)!=0)
+
+extern unsigned char* gc_page_pins;
+#define pinned_p(dummy1,dummy2) 0
+
+extern generation_index_t from_space, new_space;
+generation_index_t gc_gen_of(lispobj obj, int defaultval);
+static bool __attribute__((unused))
+from_space_p(lispobj obj)
+{
+    /* There'd be a cyclic dependency between pmrgc-impl.h and
+     * incremental-compact.h would we try to #include the latter. */
+    extern unsigned char *target_pages;
+    page_index_t page_index = find_page_index((void*)obj);
+    if (page_index == -1) return 0;
+    /* We can only move objects in or younger than the current
+     * generation, as we can't build a complete remset for older
+     * objects in a younger GC. */
+    return target_pages[page_index] && gc_gen_of(obj, 0) <= new_space;
+}
+
+static bool __attribute__((unused)) new_space_p(lispobj obj)
+{
+    gc_dcheck(compacting_p());
+    page_index_t page_index = find_page_index((void*)obj);
+    // NOTE: It is legal to access page_table at index -1,
+    // and the 'gen' of page -1 is an otherwise unused value.
+    return page_table[page_index].gen == new_space;
+}
+
 #define CHECK_COPY_PRECONDITIONS(object, nwords) \
     gc_dcheck(is_lisp_pointer(object)); \
     gc_dcheck(from_space_p(object)); \
@@ -710,38 +742,6 @@ static inline void ensure_region_closed(struct alloc_region *alloc_region,
         gc_close_region(alloc_region, page_type);
 }
 
-#define compacting_p() (from_space>=0)
-
-#define page_single_obj_p(page) ((page_table[page].type & SINGLE_OBJECT_FLAG)!=0)
-
-extern unsigned char* gc_page_pins;
-#define pinned_p(dummy1,dummy2) 0
-
-extern generation_index_t from_space, new_space;
-generation_index_t gc_gen_of(lispobj obj, int defaultval);
-static bool __attribute__((unused))
-from_space_p(lispobj obj)
-{
-    /* There'd be a cyclic dependency between pmrgc-impl.h and
-     * incremental-compact.h would we try to #include the latter. */
-    extern unsigned char *target_pages;
-    page_index_t page_index = find_page_index((void*)obj);
-    if (page_index == -1) return 0;
-    /* We can only move objects in or younger than the current
-     * generation, as we can't build a complete remset for older
-     * objects in a younger GC. */
-    return target_pages[page_index] && gc_gen_of(obj, 0) <= new_space;
-}
-
-static bool __attribute__((unused)) new_space_p(lispobj obj)
-{
-    gc_dcheck(compacting_p());
-    page_index_t page_index = find_page_index((void*)obj);
-    // NOTE: It is legal to access page_table at index -1,
-    // and the 'gen' of page -1 is an otherwise unused value.
-    return page_table[page_index].gen == new_space;
-}
-
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
 struct fixedobj_page { // 8 bytes per page
     unsigned int free_index; // index is in bytes. 4 bytes
@@ -761,8 +761,8 @@ extern struct fixedobj_page *fixedobj_pages;
 #endif
 
 extern uword_t
-walk_generation(uword_t (*proc)(lispobj*,lispobj*,uword_t),
-                generation_index_t generation, uword_t extra);
+walk_generation(uword_t (*proc)(lispobj*,lispobj*,void*),
+                generation_index_t generation, void* extra);
 
 /* The minimum heap occupancy to force more aggressive collections above. */
 #define PANIC_THRESHOLD 0.9
