@@ -15,16 +15,10 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
-#ifndef LISP_FEATURE_WIN32
-#include <sched.h>
-#endif
 #include <signal.h>
 #include <stddef.h>
 #include <errno.h>
 #include <sys/types.h>
-#ifndef LISP_FEATURE_WIN32
-#include <sys/wait.h>
-#endif
 #include "runtime.h"
 #include "validate.h"
 #include "thread.h"
@@ -32,9 +26,8 @@
 #include "target-arch-os.h"
 #include "os.h"
 #include "globals.h"
-#include "dynbind.h"
 #include "genesis/cons.h"
-#include "genesis/fdefn.h"
+#include "genesis/symbol.h"
 #include "interr.h"
 #include "genesis/sap.h"
 #include "gc.h"
@@ -170,7 +163,7 @@ void safepoint_init()
     for (i=GC_NONE; i<GC_NPHASES; ++i)
         InitializeConditionVariable(&gc_state.phase_cond[i]);
     InitializeCriticalSection(&gc_state.lock);
-#elif !defined LISP_FEATURE_X86_64
+#elif !defined LISP_FEATURE_X86_64 && !defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE
     // 64-bit already allocated a trap page when the GC card mark table was made
     os_alloc_gc_space(0, NOT_MOVABLE, GC_SAFEPOINT_PAGE_ADDR, BACKEND_PAGE_BYTES);
 #endif
@@ -465,7 +458,6 @@ thread_register_gc_trigger()
     }
 }
 
-#ifdef LISP_FEATURE_SB_SAFEPOINT
 static inline int
 thread_may_thrupt(os_context_t *ctx)
 {
@@ -564,7 +556,6 @@ check_pending_thruptions(os_context_t *ctx)
 
     return 1;
 }
-#endif
 
 int
 on_stack_p(struct thread *th, void *esp)
@@ -574,7 +565,7 @@ on_stack_p(struct thread *th, void *esp)
         < (void *)th->control_stack_end;
 }
 
-#ifndef LISP_FEATURE_WIN32
+#ifdef LISP_FEATURE_UNIX
 /* (Technically, we still allocate an altstack even on Windows.  Since
  * Windows has a contiguous stack with an automatic guard page of
  * user-configurable size instead of an alternative stack though, the
@@ -593,7 +584,7 @@ assert_on_stack(struct thread *th, void *esp)
 {
     if (on_stack_p(th, esp))
         return;
-#ifndef LISP_FEATURE_WIN32
+#ifdef LISP_FEATURE_UNIX
     if (on_altstack_p(th, esp))
         lose("thread %p: esp on altstack: %p", th, esp);
 #endif
@@ -612,7 +603,7 @@ static bool can_invoke_post_gc(struct thread* th)
 }
 
 // returns 0 if skipped, 1 otherwise
-int check_pending_gc(__attribute((unused)) os_context_t *ctx)
+int check_pending_gc()
 {
     odxprint(misc, "check_pending_gc");
     struct thread * self = get_sb_vm_thread();
@@ -720,10 +711,8 @@ void thread_in_lisp_raised(os_context_t *ctxptr)
     /* If we still need to GC, and it's not inhibited, call into
      * SUB-GC.  Phase is either GC_QUIET or GC_NONE. */
     if (check_gc_and_thruptions) {
-        check_pending_gc(ctxptr);
-#ifdef LISP_FEATURE_SB_SAFEPOINT
+        check_pending_gc();
         while(check_pending_thruptions(ctxptr));
-#endif
     }
 }
 
@@ -779,11 +768,9 @@ void thread_in_safety_transition(os_context_t *ctxptr)
             }
         }
     }
-#ifdef LISP_FEATURE_SB_SAFEPOINT
     if (was_in_alien) {
         while(check_pending_thruptions(ctxptr));
     }
-#endif
 }
 
 #ifdef LISP_FEATURE_WIN32
@@ -806,10 +793,8 @@ void thread_interrupted(os_context_t *ctxptr)
             thread_in_lisp_raised(ctxptr);
         }
     }
-    check_pending_gc(ctxptr);
-#ifdef LISP_FEATURE_SB_SAFEPOINT
+    check_pending_gc();
     while(check_pending_thruptions(ctxptr));
-#endif
 }
 #endif
 
@@ -879,7 +864,6 @@ void gc_start_the_world()
 }
 
 
-#ifdef LISP_FEATURE_SB_SAFEPOINT
 /* wake_thread(thread) -- ensure a thruption delivery to
  * `thread'. */
 
@@ -892,7 +876,7 @@ wake_thread_io(struct thread * thread)
     win32_maybe_interrupt_io(thread);
 }
 
-void wake_thread_impl(struct thread_instance *lispthread)
+static void wake_thread_impl(struct thread_instance *lispthread)
 {
     struct thread* thread = (void*)lispthread->uw_primitive_thread;
     wake_thread_io(thread);
@@ -921,7 +905,7 @@ void wake_thread_impl(struct thread_instance *lispthread)
     return;
 }
 # else
-void wake_thread_impl(struct thread_instance *lispthread)
+static void wake_thread_impl(struct thread_instance *lispthread)
 {
     struct thread *thread = (void*)lispthread->uw_primitive_thread;
     struct thread *self = get_sb_vm_thread();
@@ -980,7 +964,61 @@ void wake_thread_impl(struct thread_instance *lispthread)
     thread_sigmask(SIG_SETMASK, &oldset, 0);
 }
 #endif /* !LISP_FEATURE_WIN32 */
-#endif /* LISP_FEATURE_SB_SAFEPOINT */
+
+/* If the thread id given does not belong to a running thread (it has
+ * exited or never even existed) pthread_kill _may_ fail with ESRCH,
+ * but it is also allowed to just segfault, see
+ * <http://udrepper.livejournal.com/16844.html>.
+ *
+ * Relying on thread ids can easily backfire since ids are recycled
+ * (NPTL recycles them extremely fast) so a signal can be sent to
+ * another process if the one it was sent to exited.
+ *
+ * For these reasons, we must make sure that the thread is still alive
+ * when the pthread_kill is called and return if the thread is
+ * exiting.
+ *
+ * Note (DFL, 2011-06-22): At the time of writing, this function is only
+ * used for INTERRUPT-THREAD, hence the wake_thread special-case for
+ * Windows is OK. */
+void wake_thread(struct thread_instance* lispthread)
+{
+#ifdef LISP_FEATURE_WIN32
+    /* META: why is this comment about safepoint builds mentioning
+     * gc_stop_the_world() ? Never the twain shall meet. */
+
+    /* Kludge (on safepoint builds): At the moment, this isn't just
+     * an optimization; rather it masks the fact that
+     * gc_stop_the_world() grabs the all_threads mutex without
+     * releasing it, and since we're not using recursive pthread
+     * mutexes, the pthread_mutex_lock() around the all_threads loop
+     * would go wrong.  Why are we running interruptions while
+     * stopping the world though?  Test case is (:ASYNC-UNWIND
+     * :SPECIALS), especially with s/10/100/ in both loops. */
+
+    /* Frequent special case: resignalling to self.  The idea is
+     * that leave_region safepoint will acknowledge the signal, so
+     * there is no need to take locks, roll thread to safepoint
+     * etc. */
+    struct thread* thread = (void*)lispthread->uw_primitive_thread;
+    if (thread == get_sb_vm_thread()) {
+        sb_pthr_kill(thread, 1); // can't fail
+        check_pending_thruptions(NULL);
+        return;
+    }
+    // block_deferrables + mutex_lock looks very unnecessary here,
+    // but without them, make-target-contrib hangs in bsd-sockets.
+    sigset_t oldset;
+    block_deferrable_signals(&oldset);
+    mutex_acquire(&all_threads_lock);
+    sb_pthr_kill(thread, 1); // can't fail
+    wake_thread_impl(lispthread);
+    mutex_release(&all_threads_lock);
+    thread_sigmask(SIG_SETMASK,&oldset,0);
+#else
+    wake_thread_impl(lispthread);
+#endif
+}
 
 void* os_get_csp(struct thread* th)
 {
@@ -988,9 +1026,8 @@ void* os_get_csp(struct thread* th)
 }
 
 
-#ifndef LISP_FEATURE_WIN32
+#ifdef LISP_FEATURE_UNIX
 
-# ifdef LISP_FEATURE_SB_SAFEPOINT
 /* This is basically what 'low_level_maybe_now_maybe_later' was (which doesn't exist),
  * but with a different name, and different way of deciding to defer the signal */
 void thruption_handler(__attribute__((unused)) int signal,
@@ -1015,7 +1052,6 @@ void thruption_handler(__attribute__((unused)) int signal,
     thread_in_lisp_raised(ctx);
     csp_around_foreign_call(self) = (intptr_t) transition_sp;
 }
-# endif
 
 #ifdef LISP_FEATURE_C_STACK_IS_CONTROL_STACK
 /* Trap trampolines are in target-assem.S so that they pick up the
@@ -1057,7 +1093,7 @@ handle_safepoint_violation(os_context_t *ctx, os_vm_address_t fault_address)
     /* not a safepoint */
     return 0;
 }
-#endif /* LISP_FEATURE_WIN32 */
+#endif /* LISP_FEATURE_UNIX */
 
 void
 vodxprint_fun(const char *fmt, va_list args)
