@@ -48,20 +48,25 @@
 
 /* write_memsize_options uses a simple serialization scheme that
  * consists of one word of magic, one word indicating the size of the
- * core entry, and one word per struct field. */
+ * core entry, and one word per struct field.
+ * options == 2 => accept memsize options at runtime.
+ * options == 1 => don't accept them
+*/
 static void
-write_memsize_options(FILE *file)
+write_memsize_options(FILE *file, int options)
 {
-    core_entry_elt_t optarray[RUNTIME_OPTIONS_WORDS] = {
+    size_t size = 5 + options -1;
+    core_entry_elt_t optarray[] = {
       RUNTIME_OPTIONS_MAGIC,
-      5, // number of words in this core header entry
+      size, // number of words in this core header entry
       dynamic_space_size,
       thread_control_stack_size,
-      dynamic_values_bytes
+      dynamic_values_bytes,
+      0
     };
 
-    if (RUNTIME_OPTIONS_WORDS !=
-        fwrite(optarray, sizeof(core_entry_elt_t), RUNTIME_OPTIONS_WORDS, file)) {
+    if (size !=
+        fwrite(optarray, sizeof(core_entry_elt_t), size, file)) {
         perror("Error writing runtime options to file");
     }
 }
@@ -159,12 +164,6 @@ static long write_bytes(FILE *file, char *addr, size_t bytes,
 {
     ftell_type here, data;
 
-#ifdef LISP_FEATURE_WIN32
-    // I can't see how we'd ever attempt writing from uncommitted memory,
-    // but this is better than was was previously here (a "touch" loop over all pages)
-    os_commit_memory(addr, bytes);
-#endif
-
     fflush(file);
     here = FTELL(file);
     FSEEK(file, 0, SEEK_END);
@@ -243,8 +242,8 @@ static void unwind_binding_stack(struct thread* th)
     unbind_to_here((lispobj *)th->binding_stack_start,th);
     write_TLS(CURRENT_CATCH_BLOCK, 0, th); // If set to 0 on start, why here too?
     write_TLS(CURRENT_UNWIND_PROTECT_BLOCK, 0, th);
-    char symbol_name[] = "*SAVE-LISP-CLOBBERED-GLOBALS*";
-    lispobj* sym = find_symbol(symbol_name, get_package_by_id(PACKAGE_ID_KERNEL));
+    char symbol_name[] = "+SAVE-LISP-CLOBBERED-GLOBALS+";
+    lispobj* sym = find_symbol(symbol_name, get_package_by_id(PACKAGE_ID_SB_KERNEL));
     lispobj value;
     int i;
     if (!sym || !simple_vector_p(value = ((struct symbol*)sym)->value))
@@ -257,9 +256,22 @@ static void unwind_binding_stack(struct thread* th)
     if (verbose) printf("done]\n");
 }
 
+#ifdef LISP_FEATURE_X86_64
+static void write_static_space_constants(FILE *file)
+{
+    lispobj* ptr = (lispobj*)static_space_trailer_start;
+    unsigned int nwords = (lispobj*)STATIC_SPACE_END - ptr;
+    write_lispobj(STATIC_CONSTANTS_CORE_ENTRY_TYPE_CODE, file);
+    // +2 for the core header entry itself
+    write_lispobj(2+nwords, file);
+    /* write_lispobj(STATIC_SPACE_START, file); */
+    if (fwrite(ptr, N_WORD_BYTES, nwords, file) != nwords) perror(GENERAL_WRITE_FAILURE_MSG);
+}
+#endif
+
 bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                         bool make_executable,
-                        bool save_runtime_options,
+                        int save_runtime_options,
                         int core_compression_level)
 {
     bool verbose = !lisp_startup_options.noinform;
@@ -279,27 +291,24 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
      * all command-line arguments are available to Lisp in SB-EXT:*POSIX-ARGV*.
      * Otherwise command-line processing is performed as normal */
     if (save_runtime_options)
-        write_memsize_options(file);
+        write_memsize_options(file, save_runtime_options);
 
     int stringlen = strlen((const char *)build_id);
     int string_words = ALIGN_UP(stringlen, sizeof (core_entry_elt_t))
         / sizeof (core_entry_elt_t);
     int pad = string_words * sizeof (core_entry_elt_t) - stringlen;
-    /* Write 5 word entry header: a word for entry-type-code, the length in words,
-     * the GC enum, the address of NIL, and the string length */
+    /* Write 6 word entry header: a word for entry-type-code, the length in words,
+     * the GC enum, card table bit width, address of NIL, and string length */
     write_lispobj(BUILD_ID_CORE_ENTRY_TYPE_CODE, file);
-    write_lispobj(5 + string_words, file);
+    write_lispobj(6 + string_words, file);
 #ifdef LISP_FEATURE_GENCGC
     write_lispobj(1, file);
 #endif
 #ifdef LISP_FEATURE_MARK_REGION_GC
     write_lispobj(2, file);
 #endif
-#ifdef LISP_FEATURE_RELOCATABLE_STATIC_SPACE
-    write_lispobj(0, file);
-#else
+    write_lispobj(gc_card_table_nbits, file);
     write_lispobj(NIL, file);
-#endif
     write_lispobj(stringlen, file);
     int nwrote = fwrite(build_id, 1, stringlen, file);
     /* Write padding bytes to align to core_entry_elt_t */
@@ -314,7 +323,7 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
     /* The C runtime is theoretically position-independent so don't write out the address
      * of the unused entry sentinel. The affected elements needn't be restored on restart.
      * (Maybe add a renumbering pass in Lisp to ensure the table is 100% dense) */
-    for (i=0; i<linkage_table_count; ++i)
+    for (i=FIRST_USABLE_LINKAGE_ELT; i<linkage_table_count; ++i)
         if (linkage_space[i] == (uword_t)illegal_linkage_space_call)
             linkage_space[i] = 0;
     int nbytes = ALIGN_UP(linkage_table_count<<WORD_SHIFT, BACKEND_PAGE_BYTES);
@@ -337,6 +346,19 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
     write_lispobj(0, file); // placeholder
 
     int count = 0;
+#ifdef LISP_FEATURE_IMMOBILE_SPACE
+    /* Apparently when using MAP_32BIT on Linux, if the requested address can't be granted,
+     * sometimes the kernel won't attempt to find an alternative address that works, instead
+     * returning ENOMEM even though lack of memory is not the real issue. So now that static
+     * space is above 4GB, the next-most-problematic mapping is fixedobj space. By placing
+     * it frst in the directory, it stands the highest chance of mapping as requested. */
+    output_space(file,
+                 IMMOBILE_FIXEDOBJ_CORE_SPACE_ID,
+                 (lispobj *)FIXEDOBJ_SPACE_START,
+                 fixedobj_free_pointer,
+                 core_start_pos,
+                 core_compression_level), ++count;
+#endif
     output_space(file,
                  STATIC_CORE_SPACE_ID,
                  (lispobj *)STATIC_SPACE_START,
@@ -381,14 +403,6 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                  read_only_space_free_pointer,
                  core_start_pos,
                  core_compression_level), ++count;
-#ifdef LISP_FEATURE_IMMOBILE_SPACE
-    output_space(file,
-                 IMMOBILE_FIXEDOBJ_CORE_SPACE_ID,
-                 (lispobj *)FIXEDOBJ_SPACE_START,
-                 fixedobj_free_pointer,
-                 core_start_pos,
-                 core_compression_level), ++count;
-#endif
     // Leave this space for last! Things are easier when splitting a core into
     // code and non-code if we don't have to compensate for removal of pages.
     // i.e. if code resided between dynamic and fixedobj space, then dynamic
@@ -400,6 +414,10 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
                  core_start_pos,
                  core_compression_level), ++count;
 
+#ifdef LISP_FEATURE_X86_64
+    write_static_space_constants(file);
+#endif
+
     write_lispobj(INITIAL_FUN_CORE_ENTRY_TYPE_CODE, file);
     write_lispobj(3, file);
     write_lispobj(init_function, file);
@@ -422,8 +440,7 @@ bool save_to_filehandle(FILE *file, char *filename, lispobj init_function,
 #endif
         gc_store_corefile_ptes((struct corefile_pte*)(data + bitmapsize));
         write_lispobj(PAGE_TABLE_CORE_ENTRY_TYPE_CODE, file);
-        write_lispobj(6, file); // number of words in this core header entry
-        write_lispobj(gc_card_table_nbits, file);
+        write_lispobj(5, file); // number of words in this core header entry
         write_lispobj(next_free_page, file);
         write_lispobj(aligned_size, file);
         sword_t offset = write_bytes(file, data, aligned_size, core_start_pos,
@@ -727,15 +744,16 @@ extern void move_rospace_to_dynamic(int), prepare_readonly_space(int,int);
  */
 void
 gc_and_save(char *filename, bool prepend_runtime, bool purify,
-            bool save_runtime_options, bool compressed,
+            int save_runtime_options, bool compressed,
             int compression_level, int application_type)
 {
     // FIXME: Instead of disabling purify for static space relocation,
     // we should make r/o space read-only after fixing up pointers to
     // static space instead.
+    // BUT: On x86-64 it's ok to use the R/O space. All objects there are leaves.
 #if ((defined LISP_FEATURE_SPARC && defined LISP_FEATURE_LINUX) || \
-     (defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE))
-    /* OS says it'll give you the memory where you want, then it says
+     (defined LISP_FEATURE_RELOCATABLE_STATIC_SPACE && !defined LISP_FEATURE_X86_64))
+    /* SunOS says it'll give you the memory where you want, then it says
      * it won't map over it from the core file.  That's news to me.
      * Fragment of output from 'strace -e mmap2 src/runtime/sbcl --core output/sbcl.core':
      * ...
@@ -789,11 +807,11 @@ gc_and_save(char *filename, bool prepend_runtime, bool purify,
      * work. */
     collect_garbage(0);
 #endif
+    save_lisp_gc_iteration = 1;
     move_rospace_to_dynamic(0);
     prepare_immobile_space_for_final_gc(); // once is enough
     prepare_dynamic_space_for_final_gc(thread);
 
-    save_lisp_gc_iteration = 1;
 #ifndef LISP_FEATURE_MARK_REGION_GC
     gencgc_alloc_start_page = next_free_page;
 #endif
@@ -838,6 +856,9 @@ gc_and_save(char *filename, bool prepend_runtime, bool purify,
     // Defragment and set all objects' generations to pseudo-static
     prepare_immobile_space_for_save(verbose);
 
+#ifdef LISP_FEATURE_PPC64
+    NIL_SYMBOL_SLOTS_START[-1] = 0;
+#endif
 #ifdef LISP_FEATURE_X86_64
     untune_asm_routines_for_microarch();
 #endif
