@@ -62,7 +62,7 @@ int arch_os_thread_init(struct thread *thread) {
 #endif
 #ifdef MEMORY_SANITIZER
     extern __thread unsigned long __msan_param_tls[];
-    ((lispobj*)thread)[THREAD_MSAN_PARAM_TLS_SLOT] = (uword_t)&__msan_param_tls[0];
+    thread->msan_param_tls = (uword_t)&__msan_param_tls[0];
 #endif
     return 1;
 }
@@ -152,7 +152,7 @@ void sb_dump_mcontext(char *reason, ucontext_t* context)
     }
     sigset_tostring(&context->uc_sigmask, smallbuf, sizeof smallbuf);
     ptr += snprintf(obuf+ptr, REMAINING, "sigmask=%s\n", smallbuf);
-    write(2, obuf, ptr);
+    ignore_value(write(2, obuf, ptr));
 }
 
 os_context_register_t *
@@ -175,11 +175,16 @@ os_context_fp_addr(os_context_t *context)
     return (os_context_register_t*)&context->uc_mcontext.gregs[REG_RBP];
 }
 
-unsigned long
+unsigned int
 os_context_fp_control(os_context_t *context)
 {
-    return (uintptr_t)&context->uc_mcontext.gregs[REG_RSP];
+    return context->uc_mcontext.fpregs->mxcsr ^ (0x3F << 7);
+}
 
+void
+os_context_set_fp_control(os_context_t *context, unsigned int value)
+{
+    context->uc_mcontext.fpregs->mxcsr = value ^ (0x3F << 7);
 }
 
 os_context_register_t *
@@ -188,6 +193,18 @@ os_context_float_register_addr(os_context_t *context, int offset)
     return (os_context_register_t*)&context->uc_mcontext.fpregs->_xmm[offset];
 }
 
+os_context_register_t *
+os_context_ymm_register_addr(os_context_t *context, int offset)
+{
+#ifdef __USE_GNU
+    struct _xstate *xstate = (void*)context->uc_mcontext.fpregs;
+    return (os_context_register_t*)&(xstate->ymmh.ymmh_space[offset * 4]);
+#else
+    void *xstate = (void*)context->uc_mcontext.fpregs;
+    return (os_context_register_t*)&((char*)xstate+0x240)[offset * 16];
+#endif
+}
+
 sigset_t *
 os_context_sigmask_addr(os_context_t *context)
 {
@@ -219,21 +236,25 @@ os_flush_icache(os_vm_address_t __attribute__((unused)) address,
 #include <math.h>
 const long libm_anchor = (long)acos;
 
-#ifdef LISP_FEATURE_SW_INT_AVOIDANCE
+#if defined LISP_FEATURE_SW_INT_AVOIDANCE || defined LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
 extern void sigtrap_handler();
 extern char* vm_thread_name(struct thread*);
 extern void sigset_tostring(const sigset_t*, char*, int);
-void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
+void synchronous_trap(char* savearea, lispobj* sp_at_interrupt)
 {
     os_context_t context;
     memset(&context, 0, sizeof context);
 
     // Create the signal context from the values pushed on the stack
     // by the lisp assembly routine.
+
+#ifndef LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
+    // I don't know how to create the ucontext's representation from an XSAVE area
     context.uc_mcontext.fpregs = &context.__fpregs_mem;
     if (sizeof context.uc_mcontext.fpregs->_xmm[0].element != 16) lose("sigcontext size bug");
     memcpy(context.uc_mcontext.fpregs->_xmm[0].element, savearea, 16*16);
-    char* gprsave = savearea + 16*16;
+#endif
+    char* gprsave = savearea + 16*32; // 16 32-byte registers
     memcpy(context.uc_mcontext.gregs, gprsave, 15*8);
 
     context.uc_mcontext.gregs[REG_RSP] = (greg_t)sp_at_interrupt;
@@ -262,6 +283,10 @@ void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
             os_context_pc(&context), sp_at_interrupt, savearea,
             newmask_string); */
 
+#ifdef LISP_FEATURE_PARTIAL_SW_INT_AVOIDANCE
+    // This is for GC auto-triggers during pseudo-atomic, and nothing but
+    interrupt_handle_pending(&context);
+#else
     sigtrap_handler(0, 0, &context);
 
     if (context.uc_mcontext.gregs[REG_RSP] != (greg_t)sp_at_interrupt ||
@@ -272,6 +297,7 @@ void synchronous_trap(lispobj* sp_at_interrupt, char* savearea)
     // the return PC location that the assembly routine received.
     uword_t return_pc = context.uc_mcontext.gregs[REG_RIP];
     sp_at_interrupt[-1] = return_pc;
+#endif
 
     // act like a return-from-signal by restoring the signal mask
     // Ideally this would be performed in the asm routine only after restoring
